﻿namespace HORSSyntax
open Identifier
open HORSSyntax

type BaseSort = Ident

[<StructuredFormatDisplayAttribute("{show}")>]
type Sort =
  | SBase of BaseSort
  | SVar of SortVariable
  | SArrow of Sort * Sort

  member this.show = this.ToString ()

  override this.ToString () =
    let rec show_sty_br (b: Bool) (s: Sort) : String =
      let show_arr x y = show_sty_br true x + " -> " + show_sty_br false y
      match s with 
      | SBase n       -> n.ToString ()
      | SVar v      -> v.ToString ()
      | SArrow (u, v) -> if b then "(" + show_arr u v + ")" else show_arr u v
    show_sty_br false this

and SortVariable = 
  {
    Idnt: Ident
    Inst: Ref<Option<Sort>>
  }

  member this.show = this.ToString ()

  override this.ToString () =
    match !this.Inst with
    | None -> this.Idnt.SrcString
    | Some v -> v.ToString ()


[<StructuredFormatDisplayAttribute("{show}")>]
type SortSubstitution = 
  {
    Mappings: Map<SortVariable, Sort>
  }

  member x.show = x.ToString ()

  override x.ToString () =
    let mkBinding k v = { new Object() with member x.ToString() = sprintf "%A ~> %A" k v }
    let bindings = Map.fold (fun xs k v -> mkBinding k v :: xs) [] x.Mappings
    List.toStringWithDelims "{" ", " "}" bindings

[<StructuredFormatDisplay("{show}")>]
type SortEnv<'a> when 'a : comparison =
  {
    Env: Map<'a, Sort>
  }

  member x.show = x.ToString ()

  override x.ToString () = 
    let content = Map.fold (fun s f v -> sprintf "%s \t%A : %O\n" s f v) "" x.Env    
    sprintf "{\n %s }" content


[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module Sort =

  let o : BaseSort = Ident.mkWithStr "o"
  let d : BaseSort = Ident.mkWithStr "d"

  let tree : Sort = SBase o
  let data : Sort = SBase d

  let rec arity (s: Sort) : UInt =
    match s with
    | SBase _       -> 0u
    | SVar _      -> 0u
    | SArrow (_, v) -> 1u + arity v

  let rec order (t: Sort) : UInt =
    match t with
    | SBase _        -> 0u
    | SVar _       -> 0u
    | SArrow (t1,t2) -> max (1u + order t1) (order t2)

  let genTyVar () : SortVariable =
    {
      Idnt = Ident.mkWithSuffix "'v"
      Inst = ref None
    }

  let rec freeTyVars (s: Sort) : Set<SortVariable> =
    match s with
    | SBase _ -> Set.empty
    | SVar x -> Set.singleton x
    | SArrow (u, v) ->
        let uVars = freeTyVars u
        let vVars = freeTyVars v
        uVars + vVars

  let isVar (s: Sort) : Bool =
    match s with
    | SVar _ -> true
    | _     -> false

  let isBase (s: Sort) : Bool =
    match s with
    | SBase _ -> true
    | _      -> false

  let isArrow (s: Sort) : Bool =
    match s with
    | SArrow _ -> true
    | _       -> false

  let var (s: Sort) : SortVariable =
    match s with
    | SVar v -> v
    | _     -> failwith "Error: Not a type variable!"

  let substitute' (s: SortSubstitution) (t: Sort) : Sort =
    let rec sub t =
      Continuation.cps {
        match t with
        | SBase _ -> return t
        | SVar x ->
            match Map.tryFind x s.Mappings with
              | None -> return t
              | Some t' -> return t'
        | SArrow (u, v) ->
            let! u' = sub u
            let! v' = sub v
            return SArrow (u', v')
      }
    sub t id

  let rec substitute (s: SortSubstitution) (t: Sort) : Sort =
    match t with
    | SBase _ -> t
    | SVar x  -> 
        match Map.tryFind x s.Mappings with
        | None -> t
        | Some t' -> t'
    | SArrow (u, v) ->
        let u' = substitute s u
        let v' = substitute s v
        SArrow (u', v')

  let rec drop (n: Int) (s: Sort) : Sort =
    match n, s with
    | 0, _             -> s
    | _, SArrow (u, v) -> drop (n-1) v 
    | _, _             -> failwith "drop: not enough arguments to drop!"

//  let ofTerm ((ne, te, ve): SortCxt) (u: Term) : Sort =
//    match u with
//    | Case _ -> tree
//    | Applicative t ->
//        let headSort =
//          match t with
//          | HFun (f, _) -> Map.find f ne.Env
//          | HVar (v, _) -> Map.find v ve.Env
//          | HCst (c, _) -> Map.find c te.Env
//        drop (List.length (Term.argsInRev t)) headSort

  module Parsers =
    
    open FParsec
    open Parsers

    let pTreeSort : Parser<BaseSort> =  
      sparse {
        let! _ = pString "o"
        return o
      }

    let pTyVar : Parser<SortVariable> =
      sparse {
        let! x = Ident.Parsers.pIdent'
        return { Idnt = x; Inst = ref None }
      }

    let pSort, pSort' : Parser<Sort> * Ref<Parser<Sort>> = createParserForwardedToRef ()

    /// Atomic simple types are either base types or tyvars or simple types in parens
    let pASort : Parser<Sort> =
      let pTyVarTy = pTyVar --> SVar
      let pTreeTy  = pTreeSort --> SBase
      pTreeTy <|> pTyVarTy <|> pInParens pSort

    let pArrowSort : Parser<Sort> = 
      sparse {
        let! a = pASort
        do! pArrow
        let! b = pSort
        return SArrow (a, b)
      }

    pSort' := 
      let rec mkSort (xs: List<Sort>) : Sort =
        match xs with
        | []     -> failwith "Impossible"
        | [x]    -> x
        | y1::ys -> SArrow (y1, mkSort ys)
      sparse {
        let! xs = list pASort pArrow
        return (mkSort xs)
      }

    
  let parse (s: String) : Sort =   Parsers.parse Parsers.pSort s

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module SortSubstitution =
  
  let ofMap (m: Map<SortVariable, Sort>) : SortSubstitution =
    {
      Mappings = m
    }

  let compose (s1: SortSubstitution) (s2: SortSubstitution) : SortSubstitution =
    let s2' = Map.map (fun x t -> Sort.substitute s1 t) s2.Mappings
    let newMap = Map.fold (fun acc x t -> if Map.containsKey x s2' then acc else Map.add x t acc) s2' s1.Mappings
    ofMap newMap

  module Parsers =

    open Parsers

    let pBinding : Parser<SortVariable * Sort> =
      sparse {
        let! x = Sort.Parsers.pTyVar
        do! pTilde
        do! pRAngle
        let! s = Sort.Parsers.pSort
        return (x, s)
      }

    let pSubstitution : Parser<SortSubstitution> =
      sparse {
        let! xs = pInBraces (pList pBinding pComma)
        let sub = Map.ofList xs
        return (ofMap sub)
      }

  let parse (s: String) : SortSubstitution =
    Parsers.parse Parsers.pSubstitution s


[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module SortEnv =
  
  let ofMap (m: Map<'a, Sort>) : SortEnv<'a> =
    { Env = m }

  let substitute (sb: SortSubstitution) (e: SortEnv<'a>) : SortEnv<'a> =
    let tys = Map.map (fun _ t -> Sort.substitute sb t) e.Env
    ofMap tys

  module Parsers =
    
    open Parsers

    let pTyping (p: Parser<'a>) : Parser<'a * Sort> =
      sparse {
        let! a = p
        do! pColon
        let! s = Sort.Parsers.pSort
        return (a, s)
      }

    let pTyEnv (p: Parser<'a>) : Parser<SortEnv<'a>> =
      sparse {
        let! xs = pInBraces (list (pTyping p) pComma)
        let env = List.fold (fun m (x,y) -> Map.add x y m) Map.empty xs
        return {Env=env}
      }

