﻿namespace HORSSyntax
open HORSSyntax

exception UnifyError
exception TypeError of SynTerm * Sort * Sort

type Sorter (sc: SortEnv<SynIdent>) =

  member str.Unify (s1: Sort) (s2: Sort) : Unit =
    match s1, s2 with
    | _, _ when s1 = s2 -> ()
    | SVar x, _  -> str.UnifyVar x s2
    | _, SVar y  -> str.UnifyVar y s1
    | SArrow (u1, v1), SArrow (u2, v2) ->
        do str.Unify u1 u2
        do str.Unify v1 v2
    | _, _ -> raise UnifyError

  member str.UnifyVar (sv: SortVariable) (s: Sort) : Unit =
    match !sv.Inst with
    | None    -> str.UnifyUnboundVar sv s
    | Some s' -> str.Unify s' s

  member str.UnifyUnboundVar (sv: SortVariable) (s: Sort) : Unit =
    match s with
    | SVar x -> 
        match !x.Inst with
        | None    -> sv.Inst := Some s
        | Some s' -> str.Unify (SVar sv) s'
    | _      ->
        let s' = str.Zonk s
        let fvs = Sort.freeTyVars s'
        if Set.contains sv fvs then
          raise UnifyError
        else
          sv.Inst := Some s
  
  member str.UnifyFun (s: Sort) : Sort * Sort =
    match s with
    | SBase _ -> raise UnifyError
    | SArrow (u, v) -> (u, v)
    | SVar x  -> 
        let argSort = SVar (Sort.genTyVar ())
        let resSort = SVar (Sort.genTyVar ())
        do str.Unify s (SArrow (argSort, resSort))
        (argSort, resSort)

  member str.Check (t': SynTerm) (s: Sort) : Unit =
    let s' =
      match t' with
      | SynCase (n, x, ts) -> 
          do str.Check x Sort.data
          do List.iter (fun y -> str.Check y Sort.tree) ts
          Sort.tree
      | SynId x ->
          sc.Env.[x]
      | SynApp (u, v) ->
          let uSort = str.Infer u
          let argSort, resSort = 
            try str.UnifyFun uSort with
            | UnifyError -> 
                let argSort = SVar (Sort.genTyVar ())
                let resSort = SVar (Sort.genTyVar ())
                let expectd = SArrow (argSort, resSort)
                raise (TypeError (u, expectd, uSort))
          do str.Check v argSort
          resSort
    try str.Unify s' s with
    | UnifyError -> raise (TypeError (t', str.Zonk s, str.Zonk s'))

  member str.Infer (t: SynTerm) : Sort =
    let var = SVar (Sort.genTyVar ())
    do str.Check t var
    var

  member str.withVars (vs: List<SynIdent>) (action: Sorter -> Unit) : Unit =
    let envWithVars = 
      let mkVar () = SVar (Sort.genTyVar ())
      SortEnv.ofMap (List.fold (fun m v -> Map.add v (mkVar ()) m) sc.Env vs)
    do action (Sorter envWithVars)

  member str.Zonk (s: Sort) : Sort =
    match s with
    | SBase _ -> s
    | SArrow (u, v) -> 
        let u' = str.Zonk u
        let v' = str.Zonk v
        SArrow (u', v')
    | SVar x -> 
        match !x.Inst with
        | None -> s
        | Some s' -> 
            let s'' = str.Zonk s'
            do x.Inst := Some s''
            s''

  member str.MkArrow (xs: List<SynIdent>) : Sort =
    List.foldBack (fun v s -> SArrow (sc.Env.[v], s)) xs Sort.tree

  member str.Context = 
    
    let rec reify (s: Sort) : Sort =
      match s with
      | SBase _ -> s
      | SVar x ->
          match !x.Inst with
          | None -> s
          | Some s' -> reify s'
      | SArrow (u, v) ->
          let u' = reify u
          let v' = reify v
          SArrow (u', v')
    
    let sc' = Map.map (fun _ v -> reify v) sc.Env
    SortEnv.ofMap sc'