﻿namespace HORSSyntax
open HORSSyntax

type SynRule = 
  {
    Head: SynIdent
    Vars: List<SynIdent>
    Rhs: SynTerm
  }

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module SynRule =

  let getNonVarSyms (r: SynRule) : Set<SynIdent> =
    let addIfNotVar x s =
      if List.contains x r.Vars then s else Set.add x s
    let rec rhsSyms acc t = 
      match t with
      | SynId x -> addIfNotVar x acc
      | SynCase (n, y, xs) -> 
          let acc1 = rhsSyms acc y
          let acc2 = List.fold rhsSyms acc1 xs
          acc2
      | SynApp (u, v) ->
          let acc1 = rhsSyms acc u
          let acc2 = rhsSyms acc1 v
          acc2
    Set.add r.Head (rhsSyms Set.empty r.Rhs)

  module Parsers =

    open Parsers

    let mkRule f vs t =
      {
        Head = f
        Vars = vs
        Rhs  = t
      }

    let pRule : Parser<SynRule> =
      sparse {
        let! f  = SynIdent.Parsers.pSynIdent
        let! vs = star SynIdent.Parsers.pSynIdent
        do! pEq
        let! t  = SynTerm.Parsers.pTerm
        return mkRule f vs t
      }

  let parse (s: String) : SynRule =
    Parsers.parse Parsers.pRule s

type SynScheme =
  {
    Cxt: SortEnv<SynIdent>
    Rules: List<SynRule>
    Items: HashSet<Int>
  }

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module SynScheme = 

  let getNonVarSyms (rs: List<SynRule>) : Set<SynIdent> =
    List.fold (fun ss r -> Set.union (SynRule.getNonVarSyms r) ss) Set.empty rs

  let private mkInitCxt (ids: Set<SynIdent>) : SortEnv<SynIdent> =
    let mkVar () = SVar (Sort.genTyVar ())
    let m = Set.fold (fun m x -> Map.add x (mkVar ()) m) Map.empty ids
    SortEnv.ofMap m

  let private inferRule (s: Sorter) (r: SynRule) : Unit =
    let sortAction (str: Sorter) : Unit =
      do str.Check r.Rhs Sort.tree
      do str.Check (SynId r.Head) (str.MkArrow r.Vars)
    s.withVars r.Vars sortAction

  let init (rs: List<SynRule>) (d: HashSet<Int>) : SynScheme =
    let ids = getNonVarSyms rs
    let initCxt = mkInitCxt ids
    let str = Sorter initCxt
    do List.iter (fun r -> inferRule str r) rs
    let fnalCxt = str.Context
    {
      Cxt = fnalCxt
      Rules = rs
      Items = d
    }

  module Parsers =

    open Parsers

    let pScheme : Parser<SynScheme> =
      sparse {
        let! rs = list SynRule.Parsers.pRule pComma
        return (init rs (HashSet ()))
      }

  let parse (s: String) : SynScheme =
    Parsers.parse Parsers.pScheme s

