﻿namespace Identifier

open System.Collections.Generic

type SpanData =
  {
    Filename : String
    StartRow : Int32
    StartCol : Int32
  }

type Span =
  | Generated
  | Sourced of SpanData

[<CustomEquality>]
[<CustomComparison>]
[<StructuredFormatDisplay("{SrcString}")>]
type Ident =
  {
    UniqId    : Int32
    SrcString : String
    SrcSpan   : Span
  }

  override this.ToString () = this.SrcString

  override x.Equals that =
    let y = that :?> Ident
    x.UniqId = y.UniqId

  override this.GetHashCode() = 
    this.UniqId

  member this.CompareTo (that:obj) = this.UniqId.CompareTo (that :?> Ident).UniqId
    //Object.compHack (fun x -> x.UniqId) this that
      
  interface System.IComparable with
    member this.CompareTo that = this.CompareTo that
    

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module Ident =
  
  let number : Ref<Int32> = ref 0

  let nextId ((): Unit) : Int32 =
    begin 
      number := !number + 1;
      !number
    end

  let mkWithSuffix (str: String) : Ident =
    let uid = nextId ()
    {
      UniqId = uid
      SrcString = str + uid.ToString ()
      SrcSpan = Generated
    }

  let mkFreshFromGiven (d: Ident) : Ident =
    let uid = nextId ()
    { d with UniqId = uid }

  let mkSameWithSrcs (d: Ident) (sp: Span) : Ident =
    { d with SrcSpan = sp }

  let mkWithSrcs (str: String) (sp: Span) : Ident =
    {
      UniqId = nextId ()
      SrcString = str
      SrcSpan = sp
    }

  let mkWithStr (s: String) : Ident =
    {
      UniqId = nextId ()
      SrcString = s
      SrcSpan = Generated
    }
  
  let mkArbitrary (_: Unit) : Ident =
    let uid = nextId ()
    {
      UniqId = uid
      SrcString = "gen" + uid.ToString ()
      SrcSpan = Generated
    }

  module Parsers =
    open Parsers
    open FParsec

    let idTransTbl : Dictionary<String, Ident> = Dictionary()

    let private mkIdent : Position -> String -> Ident =
      let idTransTbl : Dictionary<String, Ident> = Dictionary()
      fun (p: Position) (s: String) ->
          let spData = 
            {
              Filename = "ThisFile"
              StartRow = int p.Line
              StartCol = int p.Column
            }
          let span = Sourced spData
          let newId = 
            match idTransTbl.TryGetValue s with
            | true,  i -> i
            | false, _ -> mkWithSrcs s span
          do idTransTbl.[s] <- newId
          newId

    let isIdChar c = isLower c
                          || isUpper c
                          || isDigit c
                          || c = '_'
                          || c = '-'
                          || c = '''

    /// Parser for identifiers prefixed by an uppercase ascii character
    let pIdentU : Parser<Ident> = 
      getPosition *> (asciiUpper **> manySatisfy isIdChar) *> getPosition
        --> fun ((p,(c,s)),p2) -> mkIdent p (c.ToString () + s)

    /// Parser for identifiers prefixed by a lowercase ascii character
    let pIdentL : Parser<Ident> = 
      getPosition *> (asciiLower **> manySatisfy isIdChar) *> getPosition
        --> fun ((p,(c,s)),p2) -> mkIdent p (c.ToString () + s)

    /// Parser for identifiers prefixed with an apostrophe character
    let pIdent' : Parser<Ident> = 
      getPosition *> (pchar ''' **> manySatisfy isIdChar) *> getPosition
        --> fun ((p,(c,s)),p2) -> mkIdent p (c.ToString () + s)

    /// Parser for identifiers without restriction
    let pIdent : Parser<Ident> = 
      getPosition *> (manySatisfy isIdChar) *> getPosition
        --> fun ((p,s),p2) -> mkIdent p s

  let parse (s: String) : Ident =
    Parsers.parse Parsers.pIdent s