﻿
// Compile with:
//  "/cygdrive/c/Program Files (x86)/Microsoft SDKs/F#/3.0/Framework/v4.0/fsc.exe" -r:../Utils/bin/Debug/Utils.dll -r:../HORSSyntax/HORSSyntax/bin/Debug/HORSSyntax.dll ModelCheckApp/HRS2APTMC.fsx


#if INTERACTIVE
#I @"../../Utils/bin/Debug/"
#I @"../../HORSSyntax/HORSSyntax/bin/Debug"
#I @"../../Identifier/Ident/bin/Debug"
#I @"../../ModelCheckLib/bin/Debug"
#r "Utils.dll"
#r "Ident.dll"
#r "HORSSyntax.dll"
#r "ModelCheckLib.dll"
#else
#endif


open HORSSyntax
open System
open ModelCheckLib

let parseHORS fn =
  try 
    Parser.pinstance fn (System.IO.File.ReadAllText fn)
  with
  | HORSParser.ParseError (str, row, col) as e ->
      eprintfn "%s:%d:%d: Parse error: %s\n" fn row col str
      exit 1
  | HORSLexer.LexerError (c, row, col) ->
      eprintfn "%s:%d:%d: Parse error: unexpected %c.\n" fn row col c
      exit 1
  | TypeError (t, s, s') -> 
      eprintfn "Sort error for term %A, expected:\n\t%A\nbut found\n\t%A\n" t s s'
      exit 1


let processOne fn =
  let inst = parseHORS fn

  let mangleSynIdent s = 
    if s.Str = "true" then { s with Str = "true_mangle" }
    elif s.Str = "false" then { s with Str = "false_mangle" }
    else s

  let rules' =
    let replaceBools (r: SynRule) : SynRule =
      let rec renameBools (t: SynTerm) : SynTerm =
        match t with
        | SynApp (u, v) -> SynApp (renameBools u, renameBools v)
        | SynCase (i, s, us) -> SynCase (i, renameBools s, List.map renameBools us)
        | SynId a -> SynId (mangleSynIdent a)

      { r with Rhs = renameBools r.Rhs
               Vars = List.map mangleSynIdent r.Vars }
    List.map replaceBools inst.HORS.Rules

  let transitions' =
    let replaceBools (tr: SynTransition) : SynTransition = { tr with Symbol = mangleSynIdent tr.Symbol }
    List.map replaceBools inst.Automaton.Transitions

  let inst = { inst with HORS = { inst.HORS with Rules = rules'
                                                 Cxt = { Env = Map.fold (fun m s v -> Map.add (mangleSynIdent s) v m) Map.empty inst.HORS.Cxt.Env } }
                         Automaton = { inst.Automaton with Transitions = transitions' } }

  let classification = Classification.ofInstance inst
  let horsc = HORSC.ofSyntax inst.HORS classification
  let calA = Automaton.ofSyntax inst.Automaton (horsc.terminals.Env |> Map.domain) classification


  let rec getNonFullyAppliedTerminals (t: TermCase) =
    let sym, args = TermCase.long t
    let rest = Set.unionMany (List.map getNonFullyAppliedTerminals args)
    match sym with
    | T a when Sort.arity horsc.terminals.Env.[a] > uint32 args.Length -> Set.add a rest
    | _ -> rest

  let generateSubstitutions (ts: Set<Terminal>) : Map<Terminal,NonTerminal> =
    let genOne (t: Terminal) =
      Identifier.Ident.mkWithStr (t.name.SrcString.Substring(0,1).ToUpper() + t.name.SrcString.Substring(1) + "_gen")
    Set.map (fun t -> t, { name = genOne t }) ts
    |> Map.ofSeq

  let rec substitute (t: TermCase) (sub: Map<Terminal,NonTerminal>) : TermCase =
    match t with
    | Sym (T x) when Map.containsKey x sub -> Sym (N sub.[x])
    | Sym _ -> t
    | App (s,t) -> App (substitute s sub, substitute t sub)

  let generateNewRules (sub: Map<Terminal,NonTerminal>) : List<NonTerminal * RewriteRuleCase> =
    let genOne (t: Terminal) (n: NonTerminal) : NonTerminal * RewriteRuleCase =
      let arity = Sort.arity horsc.terminals.Env.[t] |> int
      let freshVars = [ for i in 1 .. arity do yield {Variable.name = Identifier.Ident.mkArbitrary ()} ]
      let rhs = List.fold (fun tm v -> TermCase.App(tm, Sym (V v))) (Sym (T t)) freshVars

      n, { nt = n; vars = freshVars; rhs = rhs }

    [ for KeyValue(t, n) in sub do yield genOne t n ]


  let allNonFullyAppliedTerminals =
    [ for KeyValue(lhs,rhs) in horsc.rules do
        if rhs.Length <> 1 then failwithf "Converter assumes deterministic, but %A has multiple rules" lhs
        yield getNonFullyAppliedTerminals rhs.Head.rhs ]
    |> Set.unionMany

  let subs = generateSubstitutions allNonFullyAppliedTerminals

  let newRules = generateNewRules subs

  let subRules = Map.map (fun n rrs -> List.map (fun (rr: RewriteRuleCase) -> { rr with rhs = substitute rr.rhs subs }) rrs) horsc.rules

  let allRules = List.fold (fun m (n,rhs) -> Map.add n [rhs] m) subRules newRules

  let horsc' = { horsc with rules = allRules }

  let f = new IO.StreamWriter(IO.Path.ChangeExtension(fn,"aptmc"))
  //let f = new IO.StreamWriter(IO.Path.Combine(IO.Path.GetDirectoryName(fn),"backtranslates",IO.Path.GetFileName(fn)))
  f.WriteLine("%GRAMMAR")

  for KeyValue(nt,rrs) in horsc'.rules do
    if rrs.Length <> 1 then failwithf "Converter assumes deterministic, but %A has multiple rules" nt
    let rr = rrs.Head
    let s = sprintf "%A %s  -> %O." nt (List.toStringWithDelims "" " " "" rr.vars) rr.rhs
    f.WriteLine(s)

  f.WriteLine("\r\n%TRANSITION")
  let m = List.fold (fun m tr -> Map.updateOrDefault (tr.State,tr.Symbol) (fun l -> tr.Succs::l) [tr.Succs] m) Map.empty inst.Automaton.Transitions
  let initial = inst.Automaton.Transitions.Head.State, inst.Automaton.Transitions.Head.Symbol
    
  let prTr (q,a) qs =
    let str_of_qs =
      if List.isEmpty (List.concat qs) then "true"
      else
        [ for disjunct in qs do
            let cs =
              [ for (i,q) in disjunct do
                  yield sprintf "(%O, %O)" i q ]
            let s = String.concat @" /\ " cs
            yield if List.length qs > 1 && List.length cs > 1 then "(" + s + ")" else s
                    ]
        |> String.concat @" \/ "
    sprintf "%O %O -> %s." q a str_of_qs
    |> f.WriteLine
  prTr initial m.[initial]
  let m' = Map.remove initial m
  Map.iter prTr m'

  f.WriteLine("\r\n%PRIORITY")
  for (q,p) in inst.Automaton.Priorities do
      sprintf "%O -> %d." q p
      |> f.WriteLine

  f.Close()


#if INTERACTIVE
do if fsi.CommandLineArgs.Length >= 2 then
     do Array.iter processOne fsi.CommandLineArgs.[1..]
#else
[<EntryPoint>]
let main args = Array.iter processOne args.[1..]; 0
#endif