﻿
module Program

open BasicTypes
open ModelCheckLib
open HORSSyntax
open ModelCheckApp
open System
open System.Threading
open System.Threading.Tasks

[<EntryPoint>]
let main args =
  let extra = Options.p.Parse(args)
  if extra.Count <> 1 then
    printfn "Expected exactly one filename to process, received: '%s'" (List.toStringWithDelims "" "," "" (Seq.toList extra))
    exit 0

//  let rec myHandler (sender:obj) (args:ConsoleCancelEventArgs) =
//    printfn "Handled!"
//    //Console.CancelKeyPress.AddHandler(new ConsoleCancelEventHandler(myHandler))
//  
//  Console.CancelKeyPress.AddHandler(new ConsoleCancelEventHandler(myHandler))
//  
  let fn = extra.[0]

  let myTimer = System.Diagnostics.Stopwatch.StartNew()
  let allTimer = System.Diagnostics.Stopwatch.StartNew()
  try
    if not (System.IO.File.Exists(fn)) then
          printfn "File '%s' does not exist" fn
          exit 1
    let inst =
      try 
        Parser.pinstance fn (System.IO.File.ReadAllText fn)
      with
      | HORSParser.ParseError (str, row, col) as e ->
          eprintfn "%s:%d:%d: Parse error: %s\n" fn row col str
          exit 1
      | HORSLexer.LexerError (c, row, col) ->
          eprintfn "%s:%d:%d: Parse error: unexpected %c.\n" fn row col c
          exit 1
      | TypeError (t, s, s') -> 
          eprintfn "Sort error for term %A, expected:\n\t%A\nbut found\n\t%A\n" t s s'
          exit 1
          
    do myTimer.Restart()
    let classification = Classification.ofInstance inst
    let horsc = HORSC.ofSyntax inst.HORS classification
    let calA = Automaton.ofSyntax inst.Automaton (horsc.terminals.Env |> Map.domain) classification

    if !Cfg.verbose then
      printfn "Rule set:\n%A\nAutomaton:\n%A" horsc calA

    let btypes = Map.empty, Map.empty
    do myTimer.Stop()
    
    let property =
      (if Automaton.isConjunctive calA then "Conjunctive" else "Alternating")
      + " " +
      (if Automaton.isCoTrivial calA then "Co-trivial"
       elif Automaton.isTrivial calA then "Trivial"
       elif Automaton.isWeak calA then "Weak"
       else sprintf "%d priorities" (Map.toList calA.states |> List.map snd |> set |> Set.count))

    if !Options.latex then
      let order = horsc.nonTerminals.Env |> Map.toList |> List.map snd |> List.map Sort.order |> List.max
      let size = Map.toList horsc.rules |> List.map snd |> List.sumBy (List.sumBy (fun (rr:RewriteRuleCase) -> TermCase.size rr.rhs))
      printf "%s & %d & %d & %d & " (IO.Path.GetFileNameWithoutExtension fn) order size calA.states.Count 

    if !Options.info then
      let ntMaxArity, ntMaxOrder = Map.fold (fun (n1, n2) _ s -> let m1 = Sort.arity s in let m2 = Sort.order s in ((if m1 > n1 then m1 else n1), (if m2 > n2 then m2 else n2))) (0u, 0u) horsc.nonTerminals.Env
      let tMaxArity  = Map.fold (fun n _ s -> let m = Sort.arity s in if m > n then m else n) 0u horsc.terminals.Env
      let arity = max ntMaxArity tMaxArity
      let order = horsc.nonTerminals.Env |> Map.toList |> List.map snd |> List.map Sort.order |> List.max
      let aPriorities = Map.toList calA.states |> List.map snd |> set |> Set.count
      printf "%s Rules: %d, Order: %d, Arity: %d, Size: %d, " (IO.Path.GetFileNameWithoutExtension fn) (horsc.rules.Count) order arity !HORSParser.horsSize
      printf "States: %d, Transitions: %d, Priorities: %d, Property: " calA.states.Count calA.transitions.Count aPriorities
      printf "%s" property
      printfn "."
      exit 0
        
    if !Options.valtree <> 0 then
      let rec loop n t =
        match n with
        | 0 -> t
        | _ -> loop (n-1) (HORSC.reduce horsc t)
      loop !Options.valtree (Sym (N horsc.start))
      |> TermCase.toDot  
      |> IO.writeAll "valtree.gv"      

    let calAop = Automaton.op calA
    do myTimer.Restart()
    let result =
      if not (!Options.runsemi || !Options.rundual) then
        Options.runsemi := true; Options.rundual := true
      
      let tasks = [|
        if !Options.runsemi then
          yield Task.Factory.StartNew(fun () -> Algorithm.check horsc btypes calA) :> Task
        if !Options.rundual then
          yield Task.Factory.StartNew(fun () -> Algorithm.check horsc btypes calAop) :> Task
      |]
      
      let i = Task.WaitAny(tasks, !Options.timeout)
      if i = 0 && !Options.runsemi then
        Choice2Of3 (tasks.[i] :?> Task<Result>).Result
      elif i = 1 || not !Options.runsemi then
        Choice3Of3 (tasks.[i] :?> Task<Result>).Result
      else
        Choice1Of3 ()

    do myTimer.Stop()
 

    let checkEnv a env =
      if !Cfg.certificate then
        printfn "Witness: %A" (Yes env)
      if !Cfg.checks && (match Map.tryFind horsc.start env with
                         | None -> false
                         | Some ts when not (Set.exists (fun t -> t = IntType.atom calA.initial) ts) -> true
                         | _ -> false) then
                           printfn "ALERT: UNSOUND"
                           false
      else
        true

    if !Options.latex then printf @"%d & " myTimer.ElapsedMilliseconds
    match result with
    | Choice1Of3 _ -> if !Options.latex then printf "T" else printfn "Result: Timeout"
    
    | Choice2Of3 (Yes env) -> let ok = checkEnv calA env in if ok then (if !Options.latex then printf "Y" else printfn "Result: Yes") else (if !Options.latex then printf "I" else printfn "Result: Error")
    | Choice2Of3 (No _) -> if !Options.latex then printf "N" else (if !Cfg.certificate then printfn "Witness: All automaton choices led to error. Run with '-2' for a witness"); printfn "Result: No"
    
    | Choice3Of3 (No _) -> if !Options.latex then printf "Y" else (if !Cfg.certificate then printfn "Witness: All complement automaton choices led to error. Run with '-1' for a witness"); printfn "Result: Yes"
    | Choice3Of3 (Yes env) -> let ok = checkEnv calAop env in if ok then (if !Options.latex then printf "N" else printfn "Result: No") else (if !Options.latex then printf "I" else printfn "Result: Error")
    
    | _ -> if !Options.latex then printf "E" else printfn "Result: Error (%A)" result
    
    if !Options.latex then printfn @" & %s \\" property

    if !Cfg.timing && !Cfg.verbose then
      printfn "Total duration: %dms" allTimer.ElapsedMilliseconds
      
    0
  with
    | exc -> printfn "Exception: %s\n%s" exc.Message exc.StackTrace; exit 3


