﻿module Algorithm

open ModelCheckLib
open BasicTypes
open System.Diagnostics
open HORSSyntax

type TvMapT = CalD.TvMapT
type DerivT = CalD.DerivT

exception ErrorExc     of String
exception ExpansionExc
(*
 * This function takes the wPMRS ruleset, CalA, and current type mapping.
 * On top of that, 'j' indicates an open judgement to close.
 * The second member of the pair, 'd', must be the derivation containing 'j'.
 *)
let rec expand (r: HORSC)
               (calA: Automaton)
               (btype: BaseTypeEnv)
               (m: TvMapT)
               (cts: ChoiceTreeStore)
               badness
               (j: DerivT) : List<ChoiceTree> =

  let expand' = expand r calA btype m cts

  (* 
   * Given a type variable 'tv' (which must point to a node "App t1 t2")
   * we will add a new derivation 't2 : theta' as a child of that node
   * while setting Theta(tv) := Theta(tv) \cup theta.
   * These two modifications balance each other, keeping the derivation correct.
   *)
  let rec addDerivation (m: TvMapT) (cts: ChoiceTreeStore) (p: DerivT) (tv: TvType.TyVar) (ct: ChoiceTree) (badness: Int) (theta: TvType.T) : List<ChoiceTree> =
    let aux ((tv,j,ct): TvType.TyVar * DerivT * ChoiceTree) =

      match j.item.term, theta with
      | App (_, t2), _ ->
          
          let theta' = { theta with priority = max theta.priority j.item.itype.priority }
          TvMap.add m tv ct theta'

          let dchild = Derivation.mk t2 theta' ct (Some p) j.item.root
          RoseTree.add j dchild
          expand' badness dchild
           
      | _ -> failwith "findIntro returned a non `App'"
    in
      List.collect aux (TvMap.findIntro m cts tv ct)

  // Judgement is |- term : itype
  let term, itype, path, d = j.item.term, j.item.itype, j.item.path, j.item.root
  //printfn "Closing %A" j
  match term with
  | App (t1,_) ->
      let tv = TvMap.fresh m j
      let newtype = TvType.addLeft tv itype
      let dchild = Derivation.mk t1 newtype path (Some j) d     // t1 : a -> theta
      RoseTree.add j dchild
      expand' badness dchild
  
  | Sym (N f) -> // New |- calR(F) : itype

      Map.find f r.rules
      |> List.iteri (fun i rul -> let d' = CalD.mk calA (Some j) f (rul.vars, rul.rhs) path i itype badness
                                  do RoseTree.add d d')
      []

  | Sym (T a) ->
        let q = itype.state

        let getNth n =
          if n >= itype.stem.Length
          then failwithf "getNth failed, %d >= %d" n itype.stem.Length
          else List.nth itype.stem n

        match Map.tryFind (q, a) calA.transitions with
        | None
        | Some [] ->  Debug.printc "Latent" "Clash at %A" j
                      let clashedPaths = match cts.simpleDomination.TryGetValue(j.item.path) with
                                         | false, _ -> Set.empty
                                         | true, v -> v

                      // Mark as "bad"
                      ChoiceTree.markBadTransition cts path

                      path.Clash <- true
                      do Set.iter (fun path -> path.Clash <- true) clashedPaths
                      path :: (Set.toList clashedPaths)

        | Some [qs] ->
        
            List.collect (fun (i,q') -> addDerivation m cts j (getNth (i-1)) path 0 (TvType.mk [] q' (max calA.states.[q] calA.states.[q']))) qs
                                       
        | Some qss ->

            let path' = ChoiceTree.addChild path false None
            match cts.pathDomination.TryGetValue path with
            | false, _ -> ()
            | true, v -> for KeyValue(b,ct) in v do
                           if ct <> path then
                             let ct' = ChoiceTree.addChild ct false None
                             ChoiceTree.pathDom cts.pathDomination path' b ct'
                             ChoiceTree.simpleDom cts path' ct'
                           else
                             ChoiceTree.pathDom cts.pathDomination path' b path'
            
            let res =
              [ for i, qs in List.zip [1..qss.Length] qss do
                  let path'' = ChoiceTree.addChild path' true (Some (q,a,i))
                  match cts.pathDomination.TryGetValue path' with
                  | false, _ -> ()
                  | true, v -> for KeyValue(b,ct) in v do
                                 if ct <> path' then
                                   let ct'' = ChoiceTree.addChild ct true (Some (q,a,i))
                                   ChoiceTree.pathDom cts.pathDomination path'' b ct''
                                   ChoiceTree.simpleDom cts path'' ct''
                                 else
                                   ChoiceTree.pathDom cts.pathDomination path'' b path''

                  let badness = ChoiceTree.getBadTransitions cts (q,a,i)
                  yield! List.collect (fun (i,q') -> addDerivation m cts j (getNth (i-1)) path'' badness (TvType.mk [] q' (max calA.states.[q] calA.states.[q']))) qs ]
                
            res

  | Sym (V x) ->

      // We can avoid adding a new derivation if
      // 1. order(x) <= 1
      // 2. \Theta(\Gamma(x)) contains bla -> itype.state

      let alpha = Map.find x d.item.env
      
      // TODO: The order 1 optimisation is probably sound as long as the
      //       existing derivation has the same type and a prefix of the path
      if !Cfg.order1opt &&
         (Sort.order (Map.find x r.variables.Env) <= 0u ||
          (Sort.order (Map.find x r.variables.Env) <= 1u && Automaton.isConjunctive calA))
      then
        match List.tryFind (fun (_,(ty: TvType.T)) -> ty.state = itype.state && ty.priority = itype.priority) (TvMap.get m alpha) with
        | Some (cta,t) when ChoiceTree.ancestorOf cta path ->
            // Existing type t    = a1 -> ... -> an -> q   with path cta
            // Would create itype = b1 -> ... -> bn -> q   with path path

            Debug.printc "Bug" "Would have created %A, but %A already exists (in node %A, judgement %A)"
              (itype,path) (t,cta) d.item.id j.item

            if itype.stem.IsEmpty then
              Debug.printc "Latent" "Path %A dominates %A quietly" cta path

            if path <> cta && ChoiceTree.ancestorOf path cta then
              IO.writeAll "choicetree.gv" (ChoiceTree.toDot cts.ct)
              failwithf "ancestor check violated: %A %A" cta path

            else
              // Graft subtree of cta onto path and mark path domination
              let graft ct1 ct2 =
                let rec loop ctlist created =
                  match ctlist with
                  | [] -> ()
                  | (ct1,ct2)::rest ->
                    Debug.printc "Order1" "Grafting %A onto %A" ct1 ct2
                    if ct1.Clash then
                      ct2.Clash <- true
                      Debug.printc "Order1" "%A clashed, so filtering out %A as well" ct1 ct2
                      
                    for b in itype.stem do ChoiceTree.pathDom cts.pathDomination ct1 b ct2

                    if path <> cta then
                      ChoiceTree.simpleDom cts ct1 ct2
                      let olds = List.filter (fun ct -> not (Set.contains ct created)) ct1.Children
                      let news =
                        [ for ct in olds do yield ChoiceTree.addChild ct2 ct.Conjunctive None ]

                      loop (List.zip olds news @ rest) (Set.union (set news) created)
                    else
                      // We don't actually want to create any more nodes in this case
                      loop (List.zip ct1.Children ct1.Children @ rest) created

                loop [ct1,ct2] Set.empty

              // TV domination
              List.iter2 (TvMap.dominateBy m) itype.stem t.stem

              do graft cta path

              // Get all types of the ai
              let astypes = List.map (TvMap.get m) t.stem

              // Pull them over to the bi and add the corresponding derivations
              [ for (b,aitypes) in List.zip itype.stem astypes do
                  for ct, t in aitypes do
                    Debug.printc "Order1" "Looking up (%A,%A) in pathDomination" ct b
                    let ct' = cts.pathDomination.[ct].[b]
                    yield! addDerivation m cts j b ct' 0 t ]
            
        | _   -> addDerivation m cts j alpha path 0 itype
      else
        addDerivation m cts j alpha path 0 itype
      

let totalTimer = Stopwatch()

let toploop (r: HORSC) (a: Automaton) (btype: BaseTypeEnv) (d: CalD.T) (m: TvMapT) (cts: ChoiceTreeStore) : Result =

  let expand' = expand r a btype m cts
  let round = ref 0
  let rec loop () : Result =
    do incr round
    
    if !Cfg.verbose then printfn "Round: %d" !round

    (*********************************************)
    (* Check for Termination and Find Open Nodes *)
    (*********************************************)
    do TvMap.clearCache m
    let res =
      if !Cfg.weakopt && a.isWeak then
        CalD.nonDormantNodesWeak m d !cts.maxCompats
      else
        CalD.nonDormantNodes m d !cts.maxCompats
    
    match res with
    | CalD.Complete w -> totalTimer.Stop(); Yes w 
    | CalD.NowCheck ns ->
    
      if !Cfg.verbose then
        //printfn "=========== Round: %d" !round
        printfn "Found open nodes: %d" (List.length ns)// (List.map (fun (d:CalD.T) -> match d.item.status with CalD.Open n -> n | CalD.Closed -> -1) ns)
        printfn "  %A" (List.map (RoseTree.root >> CalD.getId) ns)
//        printfn "pathDomination: %A" [ for k in cts.pathDomination.Keys do yield k, cts.pathDomination.[k] ]
//        printfn "simpleDomination: %A" [ for k in cts.simpleDomination.Keys do yield k, cts.simpleDomination.[k] ]
//        printfn "reifycache: %A" [ for k in m.reifycache.Keys do yield k, m.reifycache.[k] ]

      let keepDecrementing ns =
        if !Cfg.badnessopt then
          not (List.exists (fun (d: CalD.T) -> d.item.status = CalD.Closed) ns)
        else
          List.exists (fun (d: CalD.T) -> not (d.item.status = CalD.Closed)) ns

      while keepDecrementing ns do
        List.iter CalD.setClosed ns

      let ns' = List.map (CalD.proof << RoseTree.root) ns
      let clashedPaths = List.collect (expand' 0) ns'
      let clashedPaths' = ChoiceTree.propagateClashes clashedPaths
      ignore clashedPaths'

      let rec allOK ct =
        let res = List.collect allOK ct.Children
        if not ct.Clash then ct :: res else res

      cts.maxCompats :=
        if !Cfg.trivialopt && Automaton.isTrivial a then
          ChoiceTree.withoutClashes cts.ct |> Set.singleton
        else
          ChoiceTree.bronKerbosch cts (set (allOK cts.ct))
        
      if Set.forall Set.isEmpty !cts.maxCompats then raise ExpansionExc

      if !round = !Cfg.maxrounds then
        Error "Reached maximum rounds"
      else
        loop ()
  in
    try
      try
        loop ()
      with
        | ExpansionExc -> No
    finally
      //printfn "Rounds: %d" !round
//      if !Cfg.verbose then
//        printfn "Size: %d" (RoseTree.size d)
//        printfn "Rounds: %d" !round
//        printfn "Model Checking time: %dms" totalTimer.ElapsedMilliseconds
        //CalD.printTypeSummary d m

     
      if !Cfg.printGraph then
        IO.writeAll "cald.gv" (CalD.toDot (RoseTree.prefix 10 d) m)
        IO.writeAll "choice.gv" (ChoiceTree.toDot cts.ct)
        //IO.writeAll "choice.txt" (sprintf "%A" ct)
        //IO.writeAll "caldc.gv" (CalD.toDotCompact (RoseTree.prefix 10 d) m)
        IO.writeAll "cald.tex" (CalD.toTex d)
        IO.writeAll "TvMap.txt" (toString m)
      ()

let check (g: HORSC) (b : BaseTypeEnv) (calA: Automaton) : Result =
  try
    let q0 = TvType.mk [] calA.initial calA.states.[calA.initial]
    let cts = ChoiceTree.emptyStore ()
    // R(S) : q0
    let calD = CalD.mk calA None g.start ([], (Map.find g.start g.rules).Head.rhs) cts.ct 0 q0 0
    cts.maxCompats := (Set.singleton (Set.singleton calD.item.proof.item.path))

    totalTimer.Restart ()

    toploop g calA b calD (TvMap.empty ()) cts
  with
    | ErrorExc s -> Error s
    | err        -> let s = sprintf "Unknown error: %s\n%s" err.Message err.StackTrace
                    Error s
