namespace ModelCheckLib
open ModelCheckLib
open HORSSyntax
open Identifier

[<StructuredFormatDisplayAttribute("{show}")>]
type Transition<'a, 'b> = 
  {
    tlhs: 'a * 'b 
    trhs: List<List<Int * 'a>>
  }

  member this.show =
    sprintf "%A %A --> %s" (fst this.tlhs) (snd this.tlhs) (List.toStringWithDelims "" " " "" this.trhs)

type Transition = Transition<Ident, Ident>

[<StructuredFormatDisplayAttribute("{show}")>]
type Transitions =
  {
    transitions: Map<State * Terminal, List<List<Int * State>>>
  }

  member this.show =
    List.toStringWithDelims "{" ", " "}" (List.map (fun (x, y) -> { tlhs=x; trhs=y }) <| Map.toList this.transitions)

[<StructuredFormatDisplayAttribute("{show}")>]
type Automaton =
  {
    alphabet: Set<Terminal>
    states: Map<State,Int>
    transitions: Map<State * Terminal, List<List<Int * State>>>
    initial: State
  }

  member this.show =
    let showAlphabet (ts: Set<Terminal>) =
      List.toStringWithDelims "    {\n      " ",\n      " "\n    }" (Set.toList ts)
    let showStates (rs: Map<State,Int>) =
      List.toStringWithDelims "    {\n      " ",\n      " "\n    }" (Map.toList rs)
    let showTransitions ts =
      List.toStringWithDelims "    {\n      " ",\n      " "\n    }" (List.map (fun (x, y) -> { tlhs=x; trhs=y }) <| Map.toList ts)
    sprintf "Automaton:\n  Alphabet:\n%s\n  States:\n%s\n  Transitions:\n%s\n  Initial state:\n    %A\n\n" (showAlphabet this.alphabet) (showStates this.states) (showTransitions this.transitions) this.initial

  member private this.isWeakMemoized = ref None

  member this.isWeak =
    // path is reversed (most recent at the head)
    let rec aux path seen (q,b) =
      if Set.contains q seen then
        Seq.forall (fun (_,b') -> b = b') (Seq.takeWhile (fun (q',_) -> q' <> q) path)
      else
        let nextStates =
          Map.filter (fun (q',_) _ -> q = q') this.transitions
          |> Map.fold (fun l _ trs -> List.concat trs :: l) []
          |> List.concat
          |> List.map snd
          |> set
          |> Set.map (fun q -> q, this.states.[q] % 2 = 0)

        let path' = (q,b) :: path
        let seen' = Set.add q seen

        Set.forall (aux path' seen') nextStates

    match !this.isWeakMemoized with
    | Some b -> b
    | None ->
        let b = aux [] Set.empty (this.initial, this.states.[this.initial] % 2 = 0)
        this.isWeakMemoized := Some b
        b


[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module Automaton =

  let op (a: Automaton) : Automaton =
    
    let cartesianProduct (sequences: List<List<'a>>) : Seq<Seq<'a>> = 
      let step acc sequence = seq { 
          for x in acc do 
          for y in sequence do 
          yield Seq.append x [y] }
      Seq.fold step (Seq.singleton Seq.empty) sequences

    let transOp (s: List<List<Int * State>>) : List<List<Int * State>> =
      let xs = cartesianProduct s
      Seq.fold (fun ys y -> (List.ofSeq y) :: ys) [] xs

    let doPair (ts: Map<State * Terminal, List<List<Int * State>>>) (q: State, c: Terminal) =
      match Map.tryFind (q, c) a.transitions with
      | None -> Map.add (q, c) [[]] ts
      | Some ss -> 
          let scs = transOp ss
          match scs with
          | [] -> ts
          | _  -> Map.add (q, c) scs ts
    {
      a with
        transitions = Set.fold doPair Map.empty (Set.product (Map.domain a.states) a.alphabet)
        states = Map.map (fun _ b -> b + 1) a.states
    }  

  let isTrivial (a: Automaton) : Bool =
    Map.forall (fun _ b -> b % 2 = 0) a.states

  let isCoTrivial (a: Automaton) : Bool =
    Map.forall (fun _ b -> b % 2 = 1) a.states

  let isConjunctive (a: Automaton) : Bool =
    Map.forall (fun (_,_) rhs -> List.length rhs = 1) a.transitions

  let isWeak (a: Automaton) : Bool = a.isWeak

  let ofSyntax (a: SynPAutomaton) (ts: Set<Terminal>) ((_,tmap,smap): Classification) : Automaton =
    let smap' s = Option.get (smap s)
    let tmap' s = Option.get (tmap s)
    let trOfSynTr tr = { tlhs = (smap' tr.State, tmap' tr.Symbol);
                         trhs = [List.map (second smap') tr.Succs] }
    let gettrhs { tlhs = _; trhs = x } = x
    let trs' = List.map trOfSynTr a.Transitions
               |> List.groupBy (fun tr1 tr2 -> tr1.tlhs = tr2.tlhs)
               |> List.fold (fun m trs -> Map.add trs.Head.tlhs (List.collect gettrhs trs) m) Map.empty
//    let synstates = seq { for tr in a.Transitions do
//                            yield  tr.State
//                            yield! List.map snd tr.Succs }
//                    |> set
    //let ts = seq { for tr in a.Transitions do yield tr.Symbol } |> Seq.map tmap' |> set

//    match Map.tryFindKey (fun _ tr -> 1 < (List.length tr)) trs' with
//    | None   -> ()
//    | Some k -> failwithf "Multiple transitions for %A" k
    let ss = List.fold (fun m (q,n) -> Map.add (smap' q) n m) Map.empty a.Priorities
    {
      alphabet = ts
      states = ss
      transitions = trs'
      initial = smap' a.Transitions.Head.State
    }
      
  module Parsers = 
  
    open Parsers

//    let pAutomaton : Parser<Automaton> = 
//      sparse {
//        let! a = SynAutomaton.Parsers.pAutomaton
//        return (ofSyntax a Classification.classifyByConvention)
//      }


//  let parse (s: String) : Automaton =
//    Parsers.parse Parsers.pAutomaton s
      
      