﻿module CalD

open BasicTypes
open ModelCheckLib
open System.IO

(*
 *  The type of CalD is a tree of pre-derivations, each labelled
 *  with a type binding F : theta. A type environment 'env' gives the
 *  type variable for each term variable. A node also comes with the open
 *  judgements that it contains. This will be a subset of those
 *  judgements that have zero children.
 *
 *)
[<StructuredFormatDisplayAttribute("{show}"); ReferenceEquality>]
type NodeT =
  {
    binding:      Binding<NonTerminal, TvType.T>
    concrete:     Option<List<T * Binding<NonTerminal,IntType> * List<Binding<NonTerminal,IntType>>>>
    proof:        Derivation.T<T>
    env:          Map<Variable, TvType.TyVar>
    status:       NodeStatus
    makesdormant: List<T>
    id:           Int
  }

  member this.show =
    sprintf "Id:%A, Type: %A\nReified: %A\nEnv: %A\nStatus:%A\nMakesdormant:%A"
      this.id
      this.binding
      this.concrete
      this.env
      this.status
      (List.map (fun (n: T) -> n.item.id) this.makesdormant)
      

and
  T = RoseTree.T<NodeT>
and
  NodeStatus =
  | Open of Int
  | Closed

let binding n = n.binding
let concrete n = n.concrete
let proof n = n.proof
let env n = n.env
let getId n = n.id

type DerivT = Derivation.T<T>
type TvMapT = TvMap<DerivT,T>

type OpenNodes =
  | NowCheck of List<T>
  | Complete of Witness

let private id = ref 0

/// Given a nonterminal 'f', a list of arguments 'xs' = x1...xn,
/// a term 't' (where calR(f) = λx1..xn.t) and an indirect type 'θ',
/// 'mk f t θ' creates a new derivation node (F : θ, |- t : state(θ)).
let mk calA p f (xs,t) ct i (θ: TvType.T) badness =
  let vs = List.fold2 (fun m x tv -> Map.add x tv m) Map.empty xs θ.stem
  let priority = calA.states.[θ.state]
  let j = Derivation.mk t (TvType.mk [] θ.state priority) ct p Unchecked.defaultof<T>
  let result = RoseTree.create { binding = { term = f; ty = θ };
                                 env = vs;
                                 concrete = None
                                 proof = j
                                 status = Open badness
                                 makesdormant = []
                                 id = !id }
  do RoseTree.update j { j.item with root = result }
  do incr id
  result
  
let size (d: T) : Int =
  let f (x: NodeT) ns = RoseTree.size x.proof + (List.sum ns)
  RoseTree.fold f d

let rm (d: T) : Unit =
  RoseTree.rm d


let hits = ref 0
let misses = ref 0

let setClosed (n: T) : Unit =
  //printfn "Closing %d" n.item.id
  match n.item.status with
  | Open 0 | Closed -> RoseTree.update n { n.item with status = Closed }
  | Open i -> RoseTree.update n { n.item with status = Open (i-1) }

let allOpenNodes _ (_: TvMapT) (calD: T) (ns: List<T>) =
  let rec loop (calD: T) =
    match calD.item.status with
    | Open _   -> [calD]
    | _ when List.contains calD ns -> []
    | _      -> List.collect loop calD.children
  loop calD

let toDotReify (calD: T) (m: TvMapT) (pi: Set<ChoiceTree>) : String =
  let rec aux (calD: T) : List<String> =
    let t = sprintf "%A: %A" calD.item.binding.term
                             (TvMap.reify m calD.item.binding.ty pi)
    let col = if not (Set.contains calD.item.proof.item.path pi) then "orange"
              else match calD.item.status with Open _ -> "red" | Closed -> "black" //| Dormant _ -> "green" | WasDormant -> "blue"
    (sprintf "\"%A\" [label=\"(D%d, %s)\" color=\"%s\" contents=\"%s\"];"
        calD.item.id
        calD.item.id t
        col
        ((sprintf "%A" calD.item.proof).Replace("\n", "\\n"))) ::
    (List.map (sprintf "\"%A\" -> \"%A\";" calD.item.id
                << getId << RoseTree.root) calD.children)
      @ (List.collect aux (List.rev calD.children))
  in
    sprintf "digraph {\ngraph [page=\"8.5,11\", aspect=1, fontname=\"Helvetica\", fontsize=\"8\", height=\"1.0\", width=\"1.0\", fixedsize=true]%s\n}\n" (aux calD |> String.concat "\n")

let inline private even n = n % 2 = 0
let inline private odd  n = n % 2 = 1

let private tryReify m (n:T) pi =
  if Set.contains n.item.proof.item.path pi then
    Some (TvMap.reifyBinding m n.item.binding pi)
  else
    None

let findWitnessWeak (m: TvMapT) (calD: T) (pi: Set<ChoiceTree>) : Witness =
  let bindings : HashSet<Binding<NonTerminal,IntType*Int>> = new HashSet<_>()

  let rec loop (ns: List<T>) witness =
    match ns with
    | [] -> witness
    | n::rest ->

        match tryReify m n pi with
        | None -> loop rest witness
        | Some b ->
          
          match n.item.status with
          | Open _ -> Debug.printc "Cert" "open node: %d" n.item.id
          | _ -> ()
          
          let witness' = Map.updateOrDefault b.term (Set.add (fst b.ty)) (Set.singleton (fst b.ty)) witness
          if even (snd b.ty) && not (bindings.Add(b)) then
            loop rest witness'
          else
            loop (rest @ n.children) witness'
      
  loop [calD] Map.empty



let nonDormantNodesWeak (m: TvMapT) (calD: T) (maxCompats: Set<Set<ChoiceTree>>) : OpenNodes =
  let bindings : HashMap<Set<ChoiceTree>,HashSet<Binding<NonTerminal,IntType>>> = new HashMap<_,_>(HashIdentity.Reference)
  let openNodes : HashMap<Set<ChoiceTree>,List<T>> = new HashMap<_,_>(HashIdentity.Reference)
  for pi in maxCompats do
    bindings.[pi] <- new HashSet<_>()
    openNodes.[pi] <- []

  let rec loop (ns: List<T*Set<Set<ChoiceTree>>>) =
    match ns with
    | [] -> ()
    | (n,pis)::rest ->
        
        match n.item.status with
        | Open _ ->
            for pi in pis do
              if Set.contains n.item.proof.item.path pi then
                openNodes.[pi] <- n :: openNodes.[pi]
              
            loop rest
              
        | _ ->
      
          let notDormant pi : Bool =
            match tryReify m n pi with
            | None -> false
            | Some b when even (snd b.ty) ->
                  // Returns true if added which implies not already seen
                  bindings.[pi].Add({ term = b.term; ty = fst b.ty })

            | _ -> true
      
          let active = Set.filter notDormant pis
      
          let nextNodes = if Set.isEmpty active then rest else rest @ [ for n in n.children -> (n, active) ]
          loop nextNodes
    
  do loop [calD,maxCompats]

  match Seq.tryPick (function KeyValue(k,[]) -> Some k | _ -> None) openNodes with
  | Some pi ->
      if !Cfg.printGraph then
        IO.writeAll "caldr.gv" (toDotReify calD m pi)
      if !Cfg.certificate then
        Complete (findWitnessWeak m calD pi)
      else
        Complete Map.empty
                                            
  | None ->
      
      let opens =
        if !Cfg.leastopensopt then
          Seq.map (function KeyValue(_,ns)-> ns) openNodes
          |> Seq.map (Seq.distinctBy (RoseTree.root >> getId))
          |> Seq.sortBy Seq.length
          |> Seq.skipWhile (fun s -> Seq.length s = 0)
          |> Seq.head
          |> Seq.toList
        else
          Seq.collect (function KeyValue(_,ns)-> ns) openNodes |> Seq.distinctBy (RoseTree.root >> getId) |> List.ofSeq

      NowCheck opens
    
let findWitness (m: TvMapT) (calD: T) (pi: Set<ChoiceTree>) : Witness =
  let extraBindings : HashSet<Binding<NonTerminal,IntType*Int>> = new HashSet<_>()

  let rec loop (ns: List<T*Set<Binding<NonTerminal,IntType*Int>>>) witness =
    match ns with
    | [] -> witness
    | (n,bindings)::rest ->

        match tryReify m n pi with
        | None -> loop rest witness
        | Some b ->
          
          match n.item.status with
          | Open _ -> Debug.printc "Cert" "open node: %d" n.item.id
          | _ -> ()
          
          let witness' = Map.updateOrDefault b.term (Set.add (fst b.ty)) (Set.singleton (fst b.ty)) witness
          if even (snd b.ty) then
            if Set.contains b bindings || extraBindings.Contains(b) then
              for b in bindings do extraBindings.Add(b) |> ignore
              loop rest witness'
            else
              let bindings' = Set.add b bindings
              loop (rest @ [ for n in n.children do yield (n, bindings') ]) witness'
          else
            let bindings' = Set.filter (fun b' -> snd b'.ty > snd b.ty) bindings
            let next = [ for n in n.children do yield (n, bindings') ]
            loop (rest @ next) witness'
      
  loop [calD, Set.empty] Map.empty


let nonDormantNodes (m: TvMapT) (calD: T) (maxCompats: Set<Set<ChoiceTree>>) : OpenNodes =
  let openNodes : HashMap<Set<ChoiceTree>,List<T>> = new HashMap<_,_>(HashIdentity.Reference)
  let extraBindings : HashMap<Set<ChoiceTree>,HashSet<Binding<NonTerminal,IntType*Int>>> = new HashMap<_,_>(HashIdentity.Reference)

  for pi in maxCompats do
    openNodes.[pi] <- []
    extraBindings.[pi] <- new HashSet<_>()

  let bindings : Map<Set<ChoiceTree>,Set<Binding<NonTerminal,IntType*Int>>> =
    Map.ofSeq (seq { for pi in maxCompats do yield pi, Set.empty })

  let rec loop (ns: List<T*Map<Set<ChoiceTree>,Set<Binding<NonTerminal,IntType*Int>>>>) =
    match ns with
    | [] -> ()
    | (n,bindings)::rest ->
        match n.item.status with
        | Open _ ->
            for KeyValue(pi,_) in bindings do
              if Set.contains n.item.proof.item.path pi then
                openNodes.[pi] <- n :: openNodes.[pi]
              
            loop rest
              
        | _ ->
          
          let bindings' =
            Map.ofSeq
              (seq {
                for KeyValue(pi,_) in bindings do
                  match tryReify m n pi with
                  | None -> ()
                  | Some b -> 
                      let m = snd b.ty

                      let seen = Set.contains b bindings.[pi] || extraBindings.[pi].Contains(b)
                      if even m then
                        if not seen then
                          yield pi, (Set.add b bindings.[pi])
                        else
                          for b in bindings.[pi] do
                            extraBindings.[pi].Add(b) |> ignore
                      else
                        yield pi, (Set.filter (fun b -> snd b.ty > m) bindings.[pi])
               })

      
          let nextNodes = if Map.isEmpty bindings' then rest else rest @ [ for n in n.children -> (n, bindings') ]
          loop nextNodes
    
  do TvMap.clearCache m
  do loop [calD,bindings]

  match Seq.tryPick (function KeyValue(k,[]) -> Some k | _ -> None) openNodes with
  | Some pi ->
      if !Cfg.printGraph then
        IO.writeAll "caldr.gv" (toDotReify calD m pi)
      if !Cfg.certificate then
        Complete (findWitness m calD pi)
      else
        Complete Map.empty
                                            
  | None ->
      
      let opens =
        if !Cfg.leastopensopt then
          Seq.map (function KeyValue(_,ns)-> ns) openNodes
          |> Seq.map (Seq.distinctBy (RoseTree.root >> getId))
          |> Seq.sortBy Seq.length
          |> Seq.skipWhile (fun s -> Seq.length s = 0)
          |> Seq.head
          |> Seq.toList
        else
          Seq.collect (function KeyValue(_,ns)-> ns) openNodes |> Seq.distinctBy (RoseTree.root >> getId) |> List.ofSeq

      NowCheck opens



(* Below this point we have debugging printers *)

let toTex (calD: T) : String =
  //let f (x: NodeT) (xs: List<String>) : String =
  //  (sprintf "$D_%d$ & %s" (incr d; !d) (Derivation.toTex x.proof)) :: xs |> String.concat "\\\\[1cm]\n"
  let f (x: NodeT) =
    let ret = (sprintf "$D_{%d} %A$\n\n%s" x.id x.env (Derivation.toTex x.proof))
    in
      fun xs -> ret :: xs |> String.concat "\\\\[2cm]\n"
  RoseTree.fold f calD

let toDot (calD: T) (_: TvMapT) : String =
  let rec aux (calD: T) : List<String> =
    let t = sprintf "%A: %A" calD.item.binding.term
                             ((*ITvMap.reify m *)calD.item.binding.ty)
    let col = match calD.item.status with Open _ -> "red" | Closed -> "black" //| Dormant _ -> "green" | WasDormant -> "blue"
    (sprintf "\"%A\" [label=\"(D%d, %s)\" color=\"%s\" contents=\"%s\"];"
        calD.item.id
        calD.item.id t
        col
        ((sprintf "%A" calD.item.proof).Replace("\n", "\\n"))) ::
    (List.map (sprintf "\"%A\" -> \"%A\";" calD.item.id
                << getId << RoseTree.root) calD.children)
      @ (List.collect aux (List.rev calD.children))
  in
    sprintf "digraph {\n%s\n}\n" (aux calD |> String.concat "\n")
  