namespace ModelCheckLib
open ModelCheckLib

type HashSet<'a> = System.Collections.Generic.HashSet<'a>
type Dictionary<'a,'b> = System.Collections.Generic.Dictionary<'a,'b>



[<CustomEquality; CustomComparison>]
[<StructuredFormatDisplayAttribute("{show}")>]
type ChoiceTree =
  {
    mutable Parent: ChoiceTree
    mutable Conjunctive: Bool
    mutable Children: List<ChoiceTree>
    mutable Clash: Bool
    Transition: Option<State * Terminal * Int>
    Id: Int
  }

  override x.Equals yobj =
    Microsoft.FSharp.Core.LanguagePrimitives.PhysicalEquality x (yobj :?> ChoiceTree)
   
  override x.GetHashCode () = Microsoft.FSharp.Core.LanguagePrimitives.PhysicalHash x

  interface System.IComparable with
    member x.CompareTo yobj =
      match yobj with
      | :? ChoiceTree as y -> Unchecked.compare x.Id y.Id
      | _ -> invalidArg "yobj" "Cannot compare elements of incompatible types"

  member this.show = this.ToString ()
  
  override this.ToString () =
    sprintf "%d" this.Id
    //sprintf "%d -> %s" this.Id (List.toStringWithDelims "[" ", " "" this.Children)


type ChoiceTreeStore =
  {
    ct: ChoiceTree
    simpleDomination: HashMap<ChoiceTree,Set<ChoiceTree>>
    pathDomination: HashMap<ChoiceTree,Map<TvType.TyVar,ChoiceTree>>
    maxCompats: Ref<Set<Set<ChoiceTree>>>
    compatTable: Dictionary<ChoiceTree*ChoiceTree,Bool>
    badTransitions: HashMap<State * Terminal * Int,Int>
  }

[<AutoOpen>]    
module ChoiceTreeActivePatterns =
  
  let (|And|Or|) (ct: ChoiceTree) =
    if ct.Conjunctive then
      And ct.Children
    else
      Or ct.Children

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module ChoiceTree =

  let private nextId = ref 0
  let private getId () = incr nextId; !nextId

  let empty () =
    let rec res = { Parent = res
                    Conjunctive = true
                    Children = []
                    Clash = false
                    Transition = None
                    Id = getId () }
    res

  let emptyStore () =
    {
      ct = empty()
      simpleDomination = new HashMap<_,_>();
      pathDomination = new HashMap<_,_>();
      badTransitions = new HashMap<_,_>();
      maxCompats = ref Set.empty
      compatTable = new Dictionary<ChoiceTree*ChoiceTree,Bool>()
    }

//  let freshChildOf p =
//     { Parent = p
//       Conjunctive = false
//       Children = []
//       Clash = false
//       Id = getId () }

  let isLeaf ct = List.isEmpty ct.Children
      
  /// `populate ct b n` causes ct to become a conjunctive node
  /// if b holds, and adds n empty children
//  let populate (ct: ChoiceTree) (b: Bool) (n: Int) =
//    if ct.Children.IsEmpty && not ct.Conjunctive then
//      ct.Children <- [ for _ in 1 .. n do yield freshChildOf ct ]
//      ct.Conjunctive <- b
//    else
//      failwithf "attempted to add children to non-leaf node: %A" ct

  let addChildren (ct: ChoiceTree) (conjunct: Bool) (cs: List<ChoiceTree>) =
    ct.Children <- cs
    ct.Conjunctive <- conjunct
    List.iter (fun ch -> ch.Parent <- ct) cs
    ct
    
    
  let markBadTransition (cts: ChoiceTreeStore) (ct: ChoiceTree) : Unit =
    match ct.Transition with
    | None -> ()
    | Some tr -> match cts.badTransitions.TryGetValue(tr) with
                 | false, _ -> cts.badTransitions.[tr] <- 1
                 | true, n  -> cts.badTransitions.[tr] <- n + 1

  let getBadTransitions (cts: ChoiceTreeStore) tr : Int =
    match cts.badTransitions.TryGetValue(tr) with
    | false, _ -> 0
    | true, n  -> n

  /// Add a child to the given choice tree, with a given
  /// conjunctiveness. Return the child.
  let addChild (ct: ChoiceTree) (conjunct: Bool) (t: Option<State * Terminal * Int>) : ChoiceTree =
    let child = { Parent = ct; Conjunctive = conjunct; Children = []; Clash = false; Transition = t; Id = getId () }
    do ct.Children <- child :: ct.Children
    child

  let pathDom (hm:HashMap<ChoiceTree,Map<TvType.TyVar,ChoiceTree>>) ct1 b ct2 =
    Debug.printc "Order1" "Path %A dominates (%A,%A)" ct1 b ct2
    match hm.TryGetValue(ct1) with
    | false, _ -> hm.Add(ct1, Map.add b ct2 Map.empty)
    | true, s -> if Map.containsKey b s then failwith "double write to pathdomination map"
                 hm.[ct1] <- Map.add b ct2 s

  let simpleDom (cts: ChoiceTreeStore) ct1 ct2 : Unit =
    match cts.simpleDomination.TryGetValue(ct1) with
    | false, _ -> cts.simpleDomination.[ct1] <- Set.singleton ct2
    | true, s -> cts.simpleDomination.[ct1] <- Set.add ct2 s

  /// Returns the list of nodes back to the root
  let private path (ct: ChoiceTree) : List<ChoiceTree> =
    let rec aux ct acc =
      if ct.Parent = ct then
        ct :: acc
      else
        aux ct.Parent (ct :: acc)
    aux ct []

  let lca ct1 ct2 : ChoiceTree =
    let rec loop last p1 p2 =
      match (p1,p2) with
      | x::xs, y::ys when x = y -> loop x xs ys
      | _,_ -> last
              
    let p1, p2 = path ct1, path ct2
    assert (p1.Head = p2.Head)
    loop p1.Head p1.Tail p2.Tail

  let propagateClashes (clashes: List<ChoiceTree>) : List<ChoiceTree> =
    let rec loop clashes =
      let clashes' =
        [ for ct in clashes do
            for ct' in ct.Children do
              if not ct'.Clash then ct'.Clash <- true; yield ct'
            if not ct.Parent.Clash && (ct.Parent.Conjunctive || List.forall (fun ct -> ct.Clash) ct.Parent.Children) then
              ct.Parent.Clash <- true; yield ct.Parent ]
        |> set
      if Set.eq clashes (Set.union clashes clashes') then 
        clashes
      else
        loop (Set.union clashes clashes')

    loop (set clashes)
    |> Set.toList

  // paths below this point that have clash = false
  let rec withoutClashes (ct: ChoiceTree) : Set<ChoiceTree> =
    let children = Set.unionMany (List.map withoutClashes ct.Children)
    if not ct.Clash then
      Set.add ct children
    else
      children

  

  /// Returns true just if ct1 and ct2 represent consistent
  /// automaton choices.
  /// This is the case if one is a descendant of the other
  /// or if their closest common ancestor is a conjunctive node.
  let compatible cts (ct1: ChoiceTree) (ct2: ChoiceTree) : Bool =
    match cts.compatTable.TryGetValue((ct1,ct2)) with
    | true, b -> b
    | false, _ ->
      let ct3 = lca ct1 ct2
      let b =
        if ct3 = ct1 || ct3 = ct2 then
          true
        else
        match ct3 with
        | And _ -> true
        | Or _  -> false
      cts.compatTable.[(ct1,ct2)] <- b
      b
    
  /// Is ct1 an ancestor of ct2?
  let ancestorOf ct1 ct2 : Bool =
    List.contains ct1 (path ct2)
  
  /// Find the maximal cliques of the presented set
  /// if the edge relation is determined by compatibility
  let bronKerbosch cts elems =
    let inline neighbours x y = compatible cts x y
    let inline minBy ct = ct.Id
    let res = new HashSet<_>()
    let emit r = ignore (res.Add(r))
      
    let rec aux r p x =  
      if Set.isEmpty p && Set.isEmpty x then
        emit r
      else
        let mutable p = p
        let mutable x = x
        
        let u =
          if p.IsEmpty then (Seq.minBy minBy x)
          elif x.IsEmpty then (Seq.minBy minBy p)
          else min (Seq.minBy minBy p) (Seq.minBy minBy x)
          
        for v in (Set.filter (fun v -> u = v || neighbours u v |> not) p) do
          let p' = Set.filter (neighbours v) (Set.remove v p)
          let x' = Set.filter (neighbours v) x

          aux (Set.add v r) p' x'
          p <- Set.remove v p
          x <- Set.add v x

    do aux Set.empty elems Set.empty
    set res
  
  let toDot (ct: ChoiceTree) : String =
    let rec loop (ct: ChoiceTree) =
      let cstr = [ for child in ct.Children do
                     yield sprintf "%d -> %d" ct.Id child.Id
                     yield loop child ]
                 |> String.concat "\n"
      sprintf "%d [label=\"%d: %s\" color=\"%s\"]\n%s" ct.Id ct.Id (if ct.Conjunctive then "And" else "Or") (if ct.Clash then "red" else "green") cstr
      
    sprintf "digraph {\n%s\n}\n" (loop ct)

  let printPathDomination = ()

  