namespace ModelCheckLib
open ModelCheckLib
open HORSSyntax
open Identifier



type Classifier<'a> = SynIdent -> Option<'a>
type Classification =Classifier<NonTerminal> * Classifier<Terminal> * Classifier<State>


[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module Classification =

  let nmap : HashMap<SynIdent,NonTerminal> = new HashMap<_,_>()
  let tmap : HashMap<SynIdent,Terminal> = new HashMap<_,_>()
  let vmap : HashMap<SynIdent,Variable> = new HashMap<_,_>()
  let smap : HashMap<SynIdent,State> = new HashMap<_,_>()

  let private spanOfSynIdent i =
    let fn, p1, p2 = i.Pos
    Sourced { Filename = fn
              StartRow = p1
              StartCol = p2 }
              
  let private newId i = Ident.mkWithSrcs i.Str (spanOfSynIdent i)
  let private sameId id i = Ident.mkSameWithSrcs id (spanOfSynIdent i)

  let mkSym (mk: Ident -> 'a) (get: 'a -> Ident) (m: Map<SynIdent,'a>) (i: SynIdent) : Map<SynIdent,'a> =
    match Map.tryFind i m with
    | Some sym -> Map.add i (mk (sameId (get sym) i)) m
    | None     -> Map.add i (mk (newId i)) m

  let nByConvention (s: SynIdent) =
    if Char.IsUpper s.Str.[0] then
      match nmap.TryGetValue s with
      | false,_ -> nmap.[s] <- { NonTerminal.name = Ident.mkWithStr s.Str }
                   Some nmap.[s]
      | true, x -> Some x
    else
      None

  let tByConvention (s: SynIdent) =
    if Char.IsLower s.Str.[0] then
      match tmap.TryGetValue s with
      | false,_ -> tmap.[s] <- { Terminal.name = Ident.mkWithStr s.Str }
                   Some tmap.[s]
      | true, x -> Some x
    else
      None

  let vByConvention (s: SynIdent) =
    if Char.IsLower s.Str.[0] then
      match vmap.TryGetValue s with
      | false,_ -> vmap.[s] <- { Variable.name = Ident.mkWithStr s.Str }
                   Some vmap.[s]
      | true, x -> Some x
    else
      None

  let sByConvention (s: SynIdent) =
    if Char.IsLower s.Str.[0] then
      match smap.TryGetValue s with
      | false,_ -> smap.[s] <- { State.name = Ident.mkWithStr s.Str }
                   Some smap.[s]
      | true, x -> Some x
    else
      None

  let classifyByConvention = nByConvention, tByConvention, sByConvention

  let ofInstance (inst: SynPInstance) : Classification =
    let mkn = mkSym (fun id -> {NonTerminal.name = id}) (fun nt -> nt.name)
    let mkv = mkSym (fun id -> {Variable.name = id}) (fun nt -> nt.name)
    let mkt = mkSym (fun id -> {Terminal.name = id}) (fun nt -> nt.name)
    let mks = mkSym (fun id -> {State.name = id}) (fun nt -> nt.name)
                     
    let nmap = List.map (fun r -> r.Head) inst.HORS.Rules
               |> List.fold mkn Map.empty 
    
    let vmap = List.collect (fun r -> r.Vars) inst.HORS.Rules
               |> List.fold mkv Map.empty 
    
    let tmap = (set (List.map (fun tr -> tr.Symbol) inst.Automaton.Transitions) ::
                List.map (fun r -> Set.difference (SynRule.getNonVarSyms r) (Map.domain nmap)) inst.HORS.Rules)
               |> List.fold (Set.fold mkt) Map.empty 

    let smap = List.append
                 (List.collect (fun tr -> tr.State :: List.map snd tr.Succs) inst.Automaton.Transitions)
                 (List.map fst inst.Automaton.Priorities)
               |> List.fold mks Map.empty

    let overlap = Set.intersect (Map.domain vmap) (Map.domain nmap)
    if not (Set.isEmpty overlap) then
      failwithf "The following identifiers are used as nonterminals and variables: '%A'" (Set.toList overlap)    
//    let overlap = Set.intersect (Map.domain vmap) (Map.domain tmap)
//    if not (Set.isEmpty overlap) then
//      failwithf "The following identifiers are used as terminals and variables: '%A'" (Set.toList overlap)
    
    (fun e -> Map.tryFind e nmap), (fun e -> Map.tryFind e tmap), (fun e -> Map.tryFind e smap)
    
    