﻿module Derivation

open BasicTypes
open ModelCheckLib



(*
 * The type of derivations is a tree of judgements of the form
 *   '|- term : itype'
 * No type environment is required as type variables have
 * replaced the variables in the term.
 *
 * The 'root' field of a judgement indicates the containing structure
 * that the proof tree is held within. It should be the same throughout
 * the tree.
 *)

[<StructuredFormatDisplayAttribute("{show}"); ReferenceEquality>]
type Judgement<'a> =
  {
    term:    TermCase;
    itype:   TvType.T;
    path:    ChoiceTree
    prev:    Option<T<'a>> // The previous judgement in the traversal
    next:    List<T<'a>> // The following judgements in the traversal
    root:    'a
  }

  member this.show =
    sprintf "|- (%d) %A : %A" this.path.Id this.term this.itype

  override this.ToString () = this.show

  member this.toTex =
    sprintf "\\vdash %s : %s" (symbolTermCaseToTex this.term) this.itype.toTex

and
  T<'a> = RoseTree.T<Judgement<'a>>

let private mkJudgement term itype path prev root =
  { term = term; itype = itype; path = path; prev = prev; next = []; root = root }

let mk term itype path prev root =
  let j = RoseTree.create (mkJudgement term itype path prev root)
  Option.iter (fun prev -> RoseTree.update prev { prev.item with next = j :: prev.item.next }) prev
  j

let rm (d: T<'a>) : Unit =
  RoseTree.rm d
  // Ensure that the previous node in the traversal
  // realises that this one has gone.
  Option.iter (fun prev -> RoseTree.update prev { prev.item with next = List.filter ((<>) d) prev.item.next }) d.item.prev

let term  j = j.term
let itype j = j.itype
let path  j = j.path
let root  j = j.root

let toTex (d: T<_>) : String =
  let rec aux (d: T<_>) : String =
    let s = (RoseTree.root d).toTex
    let ds = RoseTree.children d |> List.rev
    match ds.Length with
    | 0 -> sprintf "\\AxiomC{$%s$}\n" s
    | 1 -> sprintf "%s\\UnaryInfC{$%s$}\n" (aux ds.[0]) s
    | 2 -> sprintf "%s%s\\BinaryInfC{$%s$}\n" (aux ds.[0]) (aux ds.[1]) s
    | 3 -> sprintf "%s%s%s\\TrinaryInfC{$%s$}\n" (aux ds.[0]) (aux ds.[1]) (aux ds.[2]) s
    | _ -> sprintf "%s%s\\AxiomC{$\\cdots$}\n\\TrinaryInfC{$%s$}\n" (aux ds.[0]) (aux ds.[1]) s
  in
    sprintf "%s\\DisplayProof\n" (aux d)

