﻿namespace ModelCheckLib

open ModelCheckLib
open HORSSyntax
open Identifier

[<StructuredFormatDisplayAttribute("{show}")>]
type RewriteRuleCase =
  {
    nt:   NonTerminal
    vars: List<Variable>
    rhs:  TermCase
  }

  member this.show =
    sprintf "%A %s --> %A" this.nt (List.toStringWithDelims "" " " "" this.vars) this.rhs

[<StructuredFormatDisplayAttribute("{show}")>]
type HORSC =
  {
    terminals: SortEnv<Terminal>
    nonTerminals: SortEnv<NonTerminal>
    variables: SortEnv<Variable>
    rules: Map<NonTerminal, List<RewriteRuleCase>>
    start: NonTerminal
  }

  member this.show =
    let showTypings (m: SortEnv<'a>) =
      let tylist = Map.fold (fun xs k v -> (sprintf "%A : %A" k v) :: xs) [] m.Env
      List.toStringWithDelims "    {\n      " ",\n      " "\n    }" tylist
    let showRules (rs: List<RewriteRuleCase>) =
      List.toStringWithDelims "    {\n      " ",\n      " "\n    }" rs
    sprintf "HORSC:\n  Terminals:\n%s\n  Non-terminals:\n%s\n  Variables:\n%s\n  Rules:\n%s\n  Start:\n    %A\n\n"
      (showTypings this.terminals)
      (showTypings this.nonTerminals)
      (showTypings this.variables)
      (showRules (this.rules |> Map.toList |> List.map snd |> List.concat))
      this.start

[<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
module HORSC =

  let ofSyntax (s: SynScheme) ((nmap,tmap,_) as cl: Classification) : HORSC =
    let nmap' e = Option.get (nmap e)
    let mkv = Classification.mkSym (fun id -> {Variable.name = id}) (fun nt -> nt.name)
    let mkEnv (synmap: Classifier<'a>) : SortEnv<'a> =
      { Env = Map.fold (fun m id sort -> 
                          match synmap id with 
                          | None -> m 
                          | Some sym -> Map.add sym sort m) Map.empty s.Cxt.Env }

    let ruleOfSynRule (r: SynRule) =
      let vmap =
        let m = List.fold mkv Map.empty r.Vars
        fun e -> Map.tryFind e m
      { nt = nmap' r.Head
        vars = List.map (Option.get << vmap) r.Vars
        rhs = TermCase.ofSyntax vmap cl r.Rhs }

    let rules' = List.map ruleOfSynRule s.Rules
                 |> List.groupBy (fun r1 r2 -> r1.nt = r2.nt)
                 |> List.fold (fun m rs -> Map.add rs.Head.nt rs m) Map.empty
                 
    let rec typevars (vs,s) =
      match s, vs with
      | SArrow (s1, s2), x::xs -> (x,s1) :: typevars (xs,s2)
      | SBase _, []
      | SVar _, [] -> []
      | _,_ -> failwithf "Bad arity for a nonterminal: '%A' '%A'" s vs
      
    match Map.tryFindKey (fun _ rs -> List.length rs <> 1) rules' with
    | None -> ()
    | Some nt -> failwithf "NonTerminal %A does not have exactly one RHS: '%A'" nt rules'.[nt]

    let vmap = List.map (fun r -> rules'.[nmap' r.Head].Head.vars, s.Cxt.Env.[r.Head]) s.Rules
               |> List.collect typevars
               |> List.fold (fun m (v,s) -> Map.add v s m) Map.empty
               |> SortEnv.ofMap
    {
      terminals = mkEnv tmap
      nonTerminals = mkEnv nmap
      variables = vmap
      rules = rules'
      start = nmap' s.Rules.Head.Head
    }

  /// Reduce all outermost redexes
  let rec reduce (h: HORSC) (t: TermCase) : TermCase =
    match TermCase.long t with
    | T a, ts -> List.fold (curry App) (Sym (T a)) (List.map (reduce h) ts)
    | N f, ts -> let rr = h.rules.[f].Head //TODO: only works with deterministic schemes
                 let sub = Map.ofList (List.zip rr.vars ts)
                 TermCase.substitute rr.rhs sub
    | V _,_ -> failwith "reduce only handles ground terms"
        
(*
  let PMRSToHORSC (g: PMRS) : HORSC =
    let rulesnt nt =
      let rrs = Rule.defining nt {rules = g.rules}
      let nopat = List.forall (fun rr -> List.forall Term.isVar rr.args) rrs
      if nopat then
        List.map rrTorrc rrs
      else
        let freshvars = List.map (function (Term.Sym (T b)) -> Some (Name.genFreshVariable b.name.srcString)
                                                             | _ -> None)
                                 rrs.Head.args
        let newvars = List.map2 (fun v1 v2opt -> match v2opt with None -> (match v1 with (Term.Sym (V v)) -> v) | Some v2 -> v2) rrs.Head.args freshvars
      
        let caseargs = List.choose (function None -> None | Some v -> Some (Sym (V v))) freshvars

        let varsubs = List.map (fun rr -> rr.args) rrs
                      |> List.transpose
                      |> List.filter (List.head >> Term.isVar)
                      |> List.map (function (x::xs) -> List.fold (fun m (Term.Sym x') -> Map.add x' x m) Map.empty xs)
                      |> List.fold Term.Subst.compose Map.empty

        let cases = List.map (fun rr -> List.choose (function Term.Sym (T b) -> Some b | _ -> None) rr.args) rrs
        let mapping = List.fold2 (fun m c t -> Map.add c t m) Map.empty cases (List.map (fun (rr: RewriteRule) -> termToTermCase (Term.substitute varsubs rr.rhs)) rrs)
        let newrhs = Case (caseargs, mapping)
        [ { nt = rrs.Head.nt; vars = newvars; rhs = newrhs } ]

    let rules' = List.fold (fun m nt -> Map.add nt (rulesnt nt) m)
                           Map.empty
                           (Map.domainList g.nonTerminals)

    let typeo = Map.find g.main g.nonTerminals

    let variables' = List.map (fun rr -> rr.vars) (List.map ((flip Map.find) rules') (Map.domainList rules') |> List.concat)
                     |> List.concat
                     |> List.choose (fun v -> if List.contains v (Map.domainList g.variables) then None else Some v)
                     |> List.fold (fun m v -> Map.add v typeo m) g.variables

    {
      terminals = g.terminals
      nonTerminals = g.nonTerminals
      variables = variables'
      rules = rules'
      start = g.main
    }

  
  let baseTypes (g: PMRS) : Map<Terminal, State> * Map<State, Terminal> =
    let bs = 
      List.collect (fun rr -> rr.args) g.rules
      |> List.choose (function (Term.Sym (T t)) -> Some t | _ -> None)
      |> List.nub
      |> List.fold (fun m k -> Map.add k (Name.genFreshNamedState k.name.srcString) m) Map.empty
    in
      bs, Map.inverse bs
*)