﻿namespace ModelCheckLib
open ModelCheckLib
open HORSSyntax

[<AutoOpen>]
module IntTypeADT =

  type IntTypeImpl =
    private
    | Arrow of IntTypePSet * IntType
    | Atom  of State

  and
    [<StructuredFormatDisplayAttribute("{show}")>]
    [<CustomEquality>]
    [<CustomComparison>]
    IntType =
      private
      | One of Int * IntTypeImpl

      override x.Equals yobj =
        Microsoft.FSharp.Core.LanguagePrimitives.PhysicalEquality x (yobj :?> IntType)
       
      override x.GetHashCode () = Microsoft.FSharp.Core.LanguagePrimitives.PhysicalHash x

      member x.Id = match x with One (tag,_) -> tag

      interface System.IComparable with
        member x.CompareTo yobj = Unchecked.compare x.Id (yobj :?> IntType).Id

      member this.show =
          let rec pr (One (_,t)) =
            match t with
            | Arrow (t1, t2) ->
                let lhs = match t1.Count with
                          | 0 -> sprintf "T"
                          | 1 -> let ty, m = Seq.exactlyOne t1
                                 sprintf "(%s,%d)" (pr ty) m
                          | _ -> Seq.map (fun (x: IntType, m: Int) -> sprintf "(%s,%d)" (pr x) m) t1
                                   |> List.ofSeq 
                                   |> String.concat " ^ "
                in
                  sprintf  "%s -> %s"  lhs (pr t2)
            | Atom t   -> toString t
          in
            pr this

  and
    IntTypePSet = Set<IntType * Int>

  [<CompilationRepresentation(CompilationRepresentationFlags.ModuleSuffix)>]
  module IntType =

    let private id = ref 0
    let private getId () = incr id; !id

    let private flatEquals t1 t2 =
      match t1, t2 with
      | Atom q, Atom q' when q = q' -> true
      | Arrow (ts, t), Arrow (ts', t') -> ts = ts' && t = t'
      | _,_ -> false

    let private flatHash t =
      let combineHash x y = (x <<< 1) + y + 631 
      
      let inline hashSeq hs =
        let mutable res = 0
        for x in hs do
          res <- combineHash res (hash x)
        abs res

      match t with
      | Atom q       -> hash q
      | Arrow (ts,t) -> combineHash (hashSeq ts) (t.GetHashCode())
      
    let private flatCmp = HashIdentity.FromFunctions flatHash flatEquals
    
    let private hashedTypes = new HashMap<IntTypeImpl,IntType>(flatCmp)
    
    let private addOrReturn (t: IntTypeImpl) =
      match hashedTypes.TryGetValue t with
      | true, t' -> t'
      | false,_  ->
          lock hashedTypes (fun () -> let t' = One (getId (), t)
                                      try hashedTypes.Add(t, t'); t'
                                      with :? System.ArgumentException -> hashedTypes.[t])
      
    /// Given two priorities `m` and `n`, `subpriority m n` is `true`
    /// just if `m` is more pessimistic than `n`.
    let private subpriority (m: Int) (n: Int) =
      match m % 2 = 0, n % 2 = 0 with
      | true, true -> m <= n
      | false, false -> n <= m
      | false, true -> true
      | true, false -> false

    let rec subtype (One (_,s) : IntType) (One (_,t) : IntType) : bool =
      match s, t with
      | Atom q1, Atom q2 when q1 = q2  -> true
      | (Arrow(a, b)), (Arrow(a', b')) -> subtypeSet a' a && subtype b b'
      | Atom  _, Atom  _
      | Atom  _, Arrow _
      | Arrow _, Atom  _               -> false
    and
      subtypeSet (ss : IntTypePSet) (ts : IntTypePSet) : Bool = 
        Seq.forall (fun (x,m) -> Seq.exists (fun (y,n) -> subpriority m n && subtype y x) ss) ts    
      
    /// Is t a subtype of a member of ts, not including itself?
    let isASubtype (ts: IntTypePSet) (t: IntType, m: Int) : Bool =
      let inline isDiffSubtype (x,m) (y,n) = x <> y && subpriority m n && subtype y x
      Seq.exists (isDiffSubtype (t,m)) ts

    let rec state (One (_,t): IntType) : State =
        match t with
        | Arrow (_,t) -> state t
        | Atom q -> q

    let atom q = addOrReturn (Atom q)
      
    let arrow (ts: IntTypePSet) (t: IntType) = 
      let ts = if !Cfg.canonicalopt then (Set.filter (fun t -> not (isASubtype ts t)) ts) else ts
      addOrReturn (Arrow (ts,t))
    
  let (|Arrow|Atom|) (One (_,t): IntType) =
    match t with
    | Arrow (ts,t) -> Arrow (ts,t)
    | Atom  q      -> Atom q

    