﻿module RoseTree

[<StructuredFormatDisplayAttribute("{show}"); ReferenceEquality>]
type T<'a> =
  {
    mutable item     : 'a
    mutable children : List< T<'a> >
    mutable parent   : Option<T<'a>>
  }

  member this.show =
    let rec toStr tab t =
      sprintf "%s%A\n%s" tab t.item
        (t.children
          |> List.rev
          |> List.map (toStr (tab+" ")) 
          |> String.concat "")
    in
      toStr "" this

  member this.toDot =
    let rec aux (t: T<_>) : List<String> =
      List.map (sprintf "\"%A\" -> \"%A\";" t.item)
               (List.map (fun c -> c.item) t.children)
        @ (List.collect aux t.children)
    in
      sprintf "digraph {\n%s\n}\n" (aux this |> String.concat "\n")


let root (t: T<'a>) : 'a = t.item
let children (t: T<'a>) : List< T<'a> > = t.children

let create (x: 'a) : T<'a> =
  { item = x; children = []; parent = None }

let add (t: T<'a>) (c: T<'a>) : unit =
  t.children <- c :: t.children
  c.parent <- Some t

let rm (c: T<'a>) : Unit =
  match c.parent with
  | None -> ()
  | Some p ->
      p.children <- List.filter ((<>) c) p.children

let update (t: T<'a>) (e: 'a) : unit =
  t.item <- e

let rec fold (f: 'a -> List<'b> -> 'b) (t: T<'a>) : 'b =
  f t.item (List.map (fold f) t.children |> List.rev)

let rec iter (f: 'a -> Unit) (t: T<'a>) : Unit =
  do f t.item 
  do List.iter (iter f) (t.children |> List.rev)

let size (t: T<'a>) : Int =
  let f _ xs = 1 + (List.sum xs)
  1 + fold f t

let forall (p: 'a -> Bool) (t: T<'a>) : Bool =
  fold (fun i bs -> p i && List.fold (&&) true bs) t


let prefix (n: Int) (t: T<'a>) : T<'a> =
  let rec aux m (t: T<'a>) =
    if m = n then
      { t with children = [] }
    else
      { t with children = List.map (aux (m+1)) t.children }
  in
    aux 1 t
