﻿namespace ModelCheckLib
open ModelCheckLib

open BasicTypes

(*
 * This type represents a mapping from type variables to a list of
 * indirect types that it is bound to. It also holds a reference to
 * the point of introduction of the type variable. This is used in the
 * algorithm to find the point in a typing derivation where a type
 * variable was introduced.
 *)


type Rhs<'a,'b> =
  {
    // The open types that Theta maps this type variable to
    opentypes: List<ChoiceTree*TvType.T>
    // The judgement that caused this tv to be introduced
    intro:     'a
    // The nodes that this tv dominates. If this tv is updated,
    // then their points of introduction must also have judgements added
    dominates: List<TvType.TyVar>
  }

[<StructuredFormatDisplayAttribute("{show}")>]
type TvMap<'a,'b> =
  {
            data:  ResizeArray<Rhs<'a,'b>>;
            reifycache: HashMap<Set<ChoiceTree>,HashMap<TvType.TyVar,IntTypePSet>>;
    mutable max:   Int
  }

  member this.show =
    let ds = List.ofSeq this.data
    let mp = List.mapi (fun i d -> sprintf "%d: %A" i d.opentypes) ds
    in
      sprintf "[%s]\n" (String.concat "\n" mp)


module TvMap =

  let get m (a: TvType.TyVar) : List<ChoiceTree * TvType.T> =
    m.data.[a.value].opentypes

  let rec reify (m: TvMap<_,_>) (t: TvType.T) (cs: Set<ChoiceTree>) : IntType * Int =
    let reify t = reify m t cs

    let filterTypes ts =
      List.choose (fun (ct,ty) -> if Set.contains ct cs then Some ty else None) ts
      
    let reifySet (a: TvType.TyVar) : IntTypePSet =
      
      let compute a = List.map reify (filterTypes (get m a)) |> set
      let res =
        if !Cfg.reifyCacheOpt then
          match m.reifycache.TryGetValue(cs) with
          | true, hm -> match hm.TryGetValue a with 
                        | true,ts -> if !Cfg.checks && ts <> compute a then
                                        failwithf "reifycache failed: (%A,%A) -> '%A' <> '%A'" a cs ts (compute a)
                                     ts
                        | false, _ -> let ts = compute a
                                      hm.Add(a,ts)
                                      ts
          | false, _ -> let ts = compute a
                        let hm = new HashMap<_,_>()
                        hm.Add(a,ts)
                        m.reifycache.[cs] <- hm
                        ts
        else
          compute a
      Debug.printc "Reify" "Reifying: %A under %A, got %A" a cs res
      //Debug.printc "Reify" "cache: %A"  [ for k in m.reifycache.Keys do yield k, m.reifycache.[k] ]
      res


    let f alpha theta = IntType.arrow (reifySet alpha) theta

    let state' = IntType.atom t.state
    List.foldBack f t.stem state', t.priority

  let fresh m (d: 'a) : TvType.TyVar =
    let result = m.max
    m.max <- m.max + 1
    m.data.Add({ opentypes = [];
                 intro     = d;
                 dominates = [] })
    { value = result }
   
  /// Given 'add m a t',
  /// Adds a type 't' to the intersection represented by 'a' in the map 'm'.
  let add m (a: TvType.TyVar) (ct: ChoiceTree) (t: TvType.T) : Unit =
    Debug.printc "TvMap" "Adding %A to %A" t a
//    if List.contains t m.data.[a.value].opentypes then
//      failwith "Duplicate types in open type store"
    m.data.[a.value] <- { m.data.[a.value] with opentypes = (ct,t) :: m.data.[a.value].opentypes }

  let clearCache m : Unit =
    if !Cfg.reifyCacheOpt then
      m.reifycache.Clear()

  let rec findIntro m (cts: ChoiceTreeStore) (a: TvType.TyVar) (ct: ChoiceTree) : List<TvType.TyVar * 'a * ChoiceTree> =
    let rec aux a =
       let ct' = cts.pathDomination.[ct].[a]
       (a, m.data.[a.value].intro, ct') :: (List.collect aux m.data.[a.value].dominates)
    (a, m.data.[a.value].intro, ct) :: (List.collect aux m.data.[a.value].dominates)

  /// 'dominateBy m a b' results in the type variable 'a' being dominated by 'b',
  /// so that the intro point of 'a' will also be returned on calling
  /// 'findIntro m b'
  let dominateBy m (a: TvType.TyVar) (b: TvType.TyVar) : Unit =
    Debug.printc "Order1" "%A (%A) dominates %A (%A)" b (get m b) a (get m a)
    if a = b then
      failwithf "Error: aliasing %A to itself in %A" a m
    else
      m.data.[b.value] <- { m.data.[b.value] with dominates = a :: m.data.[b.value].dominates }

  let reifyBinding m (b: Binding<_,TvType.T>) cs : Binding<_,IntType * Int> =
    { term = b.term; ty = reify m b.ty cs }

  let empty () = {
                   data  = ResizeArray(1024)
                   reifycache = new HashMap<_,_>(HashIdentity.Reference)
                   max   = 0
                 }
