﻿module Inspect

open Microsoft.FSharp.Reflection

type Rec = { x: int }

type Tup = int * char

type Tree<'a> = Leaf of 'a
              | Node of Tree<'a> * Tree<'a>

let t = Node ((Leaf 1), Node (Leaf 5, Leaf 2))

type RecRec = { x : int; y : RecRec }

let rec recrec = { x = 6; y = recrec }

type PrimitivePrintAttribute() =
    inherit System.Attribute()

[<PrimitivePrintAttribute>]
type TestType = Banter of int

let (|Named|Array|Ptr|Param|) (typ : System.Type) =
  if typ.IsGenericType
  then Named(typ.GetGenericTypeDefinition(),typ.GetGenericArguments())
  elif typ.IsGenericParameter then Param(typ.GenericParameterPosition)
  elif not typ.HasElementType then Named(typ, [| |])
  elif typ.IsArray then Array(typ.GetElementType(),typ.GetArrayRank())
  elif typ.IsByRef then Ptr(true,typ.GetElementType())
  elif typ.IsPointer then Ptr(false,typ.GetElementType())
  else failwith "MSDN says this can't happen"

let rec formatType typ =
  match typ with
  | Named (con, [| |]) -> sprintf "%s" con.Name
  | Named (con, args) -> sprintf "%s<%s>" (con.Name.Split('`').[0]) (formatTypes args)
  | Array (arg, rank) -> sprintf "Array(%d,%s)" rank (formatType arg)
  | Ptr(true,arg) -> sprintf "%s&" (formatType arg)
  | Ptr(false,arg) -> sprintf "%s*" (formatType arg)
  | Param(pos) -> sprintf "!%d" pos
and formatTypes typs =
  String.Join(",", Array.map formatType typs) 

let inline getName (x: ^a) = (^a : (member Name : String) x)

let (|Record|_|) o =
  let ty = o.GetType()
  if FSharpType.IsRecord ty then
    Some (Array.zip (FSharpValue.GetRecordFields o) (FSharpType.GetRecordFields ty |> Array.map getName))
  else
    None

let (|Tuple|_|) o =
  let ty = o.GetType()
  if FSharpType.IsTuple ty then
    let fs = FSharpValue.GetTupleFields o
    Some (Array.zip fs (Array.map string [|1..fs.Length|]))
  else
    None

let (|Union|_|) o =
  let ty = o.GetType()
  if FSharpType.IsUnion ty then
    let fs = FSharpValue.GetUnionFields(o,ty) |> snd
    Some (Array.zip fs (Array.map string [|1..fs.Length|]))
  else
    None

let empt =
  let genericListType = typedefof<List<_>>
  let bindingFlags = System.Reflection.BindingFlags.GetProperty |||
                     System.Reflection.BindingFlags.Instance    |||
                     System.Reflection.BindingFlags.Public
  fun (o: obj) ->
      match o.GetType() with
      | t when t.IsGenericType ->
          let genericType = t.GetGenericTypeDefinition()
          if genericType = genericListType then
              t.InvokeMember("IsEmpty", bindingFlags, null, o, Array.empty) :?> bool
          else false
      | _ -> false


let empty<'a> (o: Object) =
  match o with
  | :? List<'a> as l when l.IsEmpty -> Some ()
  | _ -> None

let (|Function|_|)  o = if FSharpType.IsFunction (o.GetType()) then Some () else None
let (|Module|_|)    o = if FSharpType.IsModule   (o.GetType()) then Some () else None
let (|Exception|_|) o = if FSharpType.IsExceptionRepresentation (o.GetType()) then Some () else None
let (|Primitive|_|) o = if (o.GetType()).IsPrimitive then Some () else None
let (|Marked|_|)    o = if (o.GetType().GetCustomAttributes(typeof<PrimitivePrintAttribute>,true)).Length > 0 then Some () else None
let (|Nil|_|) (o: Object) = if empt o then Some () else None




let refcmp : System.Collections.Generic.IEqualityComparer<Object> =
  { new System.Collections.Generic.IEqualityComparer<Object> with
      member x.Equals (o1,o2) : Bool = Object.ReferenceEquals(o1,o2)
      member x.GetHashCode (o1: Object) : int = System.Runtime.CompilerServices.RuntimeHelpers.GetHashCode(o1)
  }

let getVal (d: System.Collections.Generic.Dictionary<_,_>) o = d.[o]

let printany (o: Object) : String =
  let uid = ref 0
  let fresh () = incr uid; !uid
  let seen = new System.Collections.Generic.Dictionary<Object,Int>(refcmp)
  let convert (fs : (Object * String) []) : (Int * String) [] =
    Array.map (fun (o,x) -> seen.[o],x) fs
  let rec aux (id: int) (o: Object) : String =
    //printfn "Calling aux with %d,%A" id o
    let ty = o.GetType()
    let nid = sprintf "n%d" id
    match o with
    | :? String as s -> sprintf "%s [label=\"%s\"];" nid s
    | Primitive | Nil
    | Marked      -> sprintf "%s [label=\"%A\"];" nid o
    | Record fs
    | Tuple  fs
    | Union  fs -> try
                     seen.Add(o, id)
                   with _ ->
                     eprintfn "Tried to add %A to %A" o seen
                   let olds, news = Array.partition (fun (o,_) -> seen.ContainsKey(o)) fs
                   let cids = [| for i in 1 .. news.Length do yield fresh () |]
                   sprintf "%s [label=\"%s\"];%s\n%s\n%s" nid (formatType ty)
                     (Array.map (fun (a,b) -> sprintf "%s -> n%d [label=\"%s\"];" nid a b) (convert olds) |> String.concat "\n")
                     (Array.map2 (sprintf "%s -> n%d [label=\"%s\"];" nid) cids (news |> Array.map snd) |> String.concat "\n")
                     (Array.map2 aux cids (news |> Array.map fst) |> String.concat "\n")
    | Function    -> sprintf "%s [label=\"Function: %A\"];" nid o
    | Module      -> sprintf "%s [label=\"Module\"];" nid
    | Exception   -> sprintf "%s [label=\"Exception\"];" nid
    | _           -> sprintf "%s [label=\"Unknown: %A\"];" nid o
  in
    sprintf "digraph {\nordering=out;\n%s\n}\n" (aux 0 o);


open System.IO
let viewDOTGraph (string : string) = 
   let outStream = new StreamWriter("temp.dot")
   outStream.Write string
   outStream.Close()
   
   //Create a new process to build an svg based on the graphviz
   //file we have just written.
   let graphvizProc = System.Diagnostics.Process.Start(
                           "dot",
                           "-Tsvg -o \"temp.svg\" -Kdot \"temp.dot\""
                       ) 

   graphvizProc.WaitForExit()

   //Display the svg in a viewer.
   let squiggleProc = System.Diagnostics.Process.Start(
                           "batik-squiggle.jar",
                           "temp.svg"
                       )    

   ()