﻿module List

// let pr x = sprintf "%A" x
//
// then the behaviour of this function is as:
// 
// toStringWithDelims fr sep bk [x1, ..., xn]
//   = fr + pr x1 + sep + ... + pr x(n-1) + sep + pr xn + bk
//
// toStringWithDelims is tail recursive.
//
let toStringWithDelims (fr: String) (sep: String) (bk: String) (xs: List<'a>) : String =
  let rec toSWD acc ys =
    match ys with
    | []       -> acc
    | [z]      -> sprintf "%s%A" acc z
    | y::z::zs -> toSWD (sprintf "%s%A%s" acc y sep) (z::zs)
  in
    fr + toSWD "" xs + bk

let intersperse (xs: List<'a>) (y: 'a) : List<'a> =
  let rec insp acc xs =
    match xs with
    | []         -> acc
    | [z]        -> acc @ xs
    | x1::x2::ys -> insp (acc @ [x1; y]) (x2::ys)
  insp [] xs

let contains (x: 'a) (ys: List<'a>) : Bool when 'a : equality =
  Option.isSome (List.tryFind (fun y -> y = x) ys)

let rec allDifferent (xs: List<'a>) : Bool when 'a : equality =
  match xs with
  | []    -> true
  | x::xs -> not (contains x xs) && allDifferent xs

let nub (xs: List<'a>) : List<'a> =
  let rec nub' acc ys =
    match ys with
    | []    -> []
    | z::zs -> if contains z acc then nub' acc zs else z::nub' (z::acc) zs
  in
    nub' [] xs

let remove (x: 'a) (xs: List<'a>) : List<'a> =
  List.filter (fun y -> x <> y) xs

/// Tail recursive map-accumulate function
let mapAccum (f: 'a -> 'b -> 'a * 'c) (a: 'a) (xs: List<'b>) : List<'c> =
  let rec mapA acc ys =
    Continuation.cps {
      match ys with
      | []    -> return []
      | z::zs -> 
          let  acc', z' = f acc z
          let! zs' = mapA acc' zs
          return z'::zs'
    }
  mapA a xs id

let productWith (p: 'a -> 'b -> 'c) (xs: List<'a>) (ys: List<'b>) : List<'c> =
  List.fold (fun ps x -> List.fold (fun s y -> p x y::s) ps ys) [] xs

let product (xs: List<'a>) (ys: List<'b>) : List<'a * 'b> =
  productWith (fun x y -> (x, y)) xs ys

let productL (xss: List<List<'a>>) : List<List<'a>> =
  let product' (ys: List<'a>) (xss: List<List<'a>>) : List<List<'a>> =
    [ for xs in xss do yield! List.map (fun y -> y::xs) ys ]
  List.foldBack product' xss [[]]

let findAllIndexes (y: 'a) (xs: List<'a>) : List<UInt32> when 'a : comparison =
  snd <| List.fold (fun (n, inds) x -> (n+1u, if x = y then n::inds else inds)) (0u, []) xs

let groupBy (eq: 'a -> 'a -> Bool) (xs: List<'a>) : List<List<'a>> =
  let rec group grpd rest =
    match rest with
    | []    -> grpd
    | y::ys ->  
        let grp, others = List.partition (eq y) ys
        group ((y::grp)::grpd) others
  group [] xs

///Given list of 'rows', returns list of 'columns' 
let rec transpose lst =
  match lst with
  | (_::_)::_ -> List.map List.head lst :: transpose (List.map List.tail lst)
  | _         -> []
