﻿module Map

let union (m: Map<'a,'b>) (n: Map<'a,'b>) : Map<'a, 'b> =
  Map.fold (fun s k t -> Map.add k t s) m n

let singleton (k: 'a) (x: 'b) : Map<'a, 'b> =
  Map.add k x Map.empty

let update (k: 'a) (f: 'b -> 'b) (m: Map<'a, 'b>) : Map<'a, 'b> =
  let existing = Map.find k m
  Map.add k (f existing) m

let updateOrDefault (k: 'a) (f: 'b -> 'b) (d: 'b) (m: Map<'a, 'b>) : Map<'a, 'b> =
  match Map.tryFind k m with
  | None -> Map.add k d m
  | Some b -> Map.add k (f b) m

let inverse (m: Map<'a, 'b>) : Map<'b, 'a> =
  Map.fold (fun im k v -> Map.add v k im) Map.empty m

let domainList (m: Map<'a, 'b>) : List<'a> =
  Map.fold (fun xs k _ -> k::xs) [] m

let domain (m: Map<'a, 'b>) : Set<'a> =
  Map.fold (fun xs k _ -> Set.add k xs) Set.empty m

/// 'init xs e' yields a map m such that dom(m) = xs
/// and forall x in xs, m(x) = e
let init (xs: List<'a>) (e: 'b) : Map<'a,'b> =
  List.fold (fun m k -> Map.add k e m) Map.empty xs

/// `eq eqb m n' is `true' if maps `m' and `n' are pointwise equal in terms of the supplied
/// equality function `eqb' which acts on the elements of the range
let eq (eqb: 'b -> 'b -> Bool) (m: Map<'a, 'b>) (n: Map<'a, 'b>) : Bool =

  let checkPoints (v: 'b) (wopt: Option<'b>) : Bool =
    match wopt with
    | None -> false
    | Some x -> eqb x v

  Map.forall (fun k v -> checkPoints v (Map.tryFind k n)) m && Map.forall (fun k v -> checkPoints v (Map.tryFind k m)) n

type MapComprehension() =
  
  member this.For (s: seq<'k * 'v>) (f: 'k * 'v -> Map<'k, 'v>) : Map<'k, 'v> =
    Seq.fold (fun m p -> union m (f p)) Map.empty s

  member this.Yield ((k, v): 'k * 'v) : Map<'k, 'v> =
    singleton k v

  member this.YieldFrom (m: Map<'k, 'v>) : Map<'k, 'v> =
    m

let mapComp = new MapComprehension()
