﻿module Loader

#r "Microsoft.Build"
#r "Microsoft.Build.Engine.dll"
#r "Microsoft.Build.Framework.dll"
#r "Microsoft.Build.Tasks.v4.0.dll"
#r "Microsoft.Build.Utilities.v4.0.dll"

open System
open System.IO
open Microsoft.Build.Tasks
open Microsoft.Build.Utilities
open Microsoft.Build.Framework
open Microsoft.Build.BuildEngine  
open Microsoft.Build

let eqHack (f: 'a -> 'b) (x: 'a) (yobj: Object) : Boolean =
  match yobj with
  | :? 'a as y -> f x = f y
  | _          -> false

let compHack (f: 'a -> 'b) (x: 'a) (yobj: Object) : Int32 =
  match yobj with
  | :? 'a as y -> compare (f x) (f y)
  | _          -> invalidArg "yobj" "Cannot compare elements of incompatible types"

let contains (x: 'a) (ys: List<'a>) : bool when 'a : equality =
  Option.isSome (List.tryFind (fun y -> y = x) ys)

let nubBy (f: 'a -> 'b) (xs: List<'a>) : List<'a> =
  let rec nub' acc ys =
    match ys with
    | []    -> []
    | z::zs -> if contains (f z) acc then nub' acc zs else z::nub' (f z::acc) zs
  in
    nub' [] xs

let preloads =
  [
    "System"
    "System.Core"
    "mscorlib"
    "FSharp.Core"
    "System.Numerics"
  ]

type Digraph<'n> when 'n : comparison = 
  Map<'n, Set<'n>> 

module Digraph =
  
  let addNode (n: 'n) (g: Digraph<'n>) : Digraph<'n> =
    match Map.tryFind n g with
    | None -> Map.add n Set.empty g
    | Some _ -> g

  let addEdge ((n1, n2): 'n * 'n) (g: Digraph<'n>) : Digraph<'n> =
    let g' = 
      match Map.tryFind n2 g with
      | None -> addNode n2 g
      | Some _ -> g
    match Map.tryFind n1 g with
    | None -> Map.add n1 (Set.singleton n2) g'
    | Some ns -> Map.add n1 (Set.add n2 ns) g'

  let nodes (g: Digraph<'n>) : List<'n> =
    Map.fold (fun xs k _ -> k::xs) [] g

  let roots (g: Digraph<'n>) : List<'n> =
    List.filter (fun n -> not (Map.exists (fun _ v -> Set.contains n v) g)) (nodes g)

  let topSort (h: Digraph<'n>) : List<'n> =
    let rec dfs (g: Digraph<'n>, order: List<'n>, rts: List<'n>) : List<'n> =
      if List.isEmpty rts then
        order
      else
        let n = List.head rts
        let children = Map.find n g
        let order' = n::order
        let g' = Map.remove n g
        let rts' = roots g'
        dfs (g', order', rts')
    dfs (h, [], roots h)

type Item(p: ITaskItem) = 
  
  member this.TaskItem : ITaskItem = p

  override this.Equals (obj: Object) : bool =
    eqHack (fun (p:Item) -> p.TaskItem.ItemSpec) this obj

  override this.GetHashCode () = 
    hash this.TaskItem.ItemSpec
  
  interface System.IComparable with 
    member this.CompareTo (obj: Object) =
      compHack (fun (p:Item) -> p.TaskItem.ItemSpec) this obj

let toTaskItem (i: Item) : ITaskItem = i.TaskItem
let toItem (i: ITaskItem) : Item = Item(i)

let fwRefs = ToolLocationHelper.GetPathToDotNetFrameworkReferenceAssemblies TargetDotNetFrameworkVersion.Version40
printf "%s" fwRefs
let rar = ResolveAssemblyReference ()
rar.Silent <- true
rar.SearchPaths <- [| "{RawFileName}" |]
//rar.TargetProcessorArchitecture <- "MSIL"
//rar.TargetFrameworkDirectories <- [|fwRefs|]

let resolve (r: Item) : Item * array<Item> =
  rar.Assemblies <- [|toTaskItem r|]
  printf "Resolving: %O\n" r.TaskItem.ItemSpec
  ignore (rar.Execute ())
  let resolvedItem =
    printf "Resolved files: %A" rar.ResolvedFiles
    if Array.length rar.ResolvedFiles <> 1 then
      failwithf "Could not resolve dependency: %s" r.TaskItem.ItemSpec
    else
      rar.ResolvedFiles.[0]
  let resolvedDeps = rar.ResolvedDependencyFiles
  (toItem resolvedItem, Array.map toItem resolvedDeps)

let allFsProjFiles dir (fsprojfile: string) =
  let aux (g: Digraph<string>) (proj: string) : Digraph<string> =
    let d     = Execution.ProjectInstance proj
    let prefs = d.GetItems "ProjectReference"
    let cs    = Seq.map (fun (c: Execution.ProjectItemInstance) -> Path.Combine (dir, c.EvaluatedInclude)) prefs
    Seq.fold (fun g c -> Digraph.addEdge (proj,c) g) g cs
  let rec fixpoint (g: Digraph<string>) : Digraph<string> =
    let ns = Digraph.nodes g
    let g' = List.fold aux g ns
    if g = g' then g else fixpoint g'
  let projToLoad (proj: string) : seq<string> =
    let d     = Execution.ProjectInstance proj
    let dir   = Path.GetDirectoryName proj
    let files = d.GetItems "Compile"
    Seq.map (fun (fsFile:Execution.ProjectItemInstance) -> "#load \"" + (Path.Combine(dir,fsFile.EvaluatedInclude)) + "\"") files
  in
    let g = fixpoint (Digraph.addNode fsprojfile Map.empty)
    Digraph.topSort g
    |> Seq.collect projToLoad

let mkGraph (seeds: seq<ITaskItem>) : Digraph<Item> =
  let processNode (g: Digraph<Item>) (n: Item) : Digraph<Item> =
    let resItem, resDeps = resolve n
    let g' = Digraph.addNode resItem (Map.remove n g)
    Seq.fold (fun h c -> Digraph.addEdge (resItem, c) h) g' resDeps
  let rec fixpoint (g: Digraph<Item>) : Digraph<Item> =
    let ns = Digraph.nodes g
    let g' = List.fold processNode g ns
    if g = g' then g else fixpoint g'
  fixpoint (Seq.fold (fun g s -> Digraph.addNode s g) Map.empty (Seq.map toItem seeds))

let spec (i: Item) : String * Option<string * string> =
  if i.TaskItem.GetMetadata "CopyLocal"  = "true" then 
    let src  = i.TaskItem.GetMetadata "FullPath"
    let dest = System.IO.Path.GetFileName src
    (dest, Some (src, dest))
  elif i.TaskItem.GetMetadata "ResolvedFrom" = "{GAC}" then
    (i.TaskItem.GetMetadata "Filename", None)
  else
    (i.TaskItem.GetMetadata "FullPath", None)

let cpDll (s: string, o: Option<string * string>) : Unit =
  match o with
  | None -> ()
  | Some (src, dest) ->
      System.IO.File.Copy(src, dest, true)    

let projRefToItem (pr: Execution.ProjectItemInstance) : TaskItem =
  let proj   = Execution.ProjectInstance pr.EvaluatedInclude
  let outDir = System.IO.Path.Combine (proj.Directory, proj.GetPropertyValue("OutputPath"))
  let fname  = proj.GetPropertyValue "Name" + ".dll"
  let path   = System.IO.Path.Combine (outDir, fname)
  Utilities.TaskItem(path)

let load (sd: String) : Unit =

  let cd = System.Environment.CurrentDirectory
  printf "%s\n" cd
  do System.Environment.CurrentDirectory <- sd
  printf "%s\n" sd
  printf "%s\n" __SOURCE_DIRECTORY__

  let projFile = Directory.GetFiles(sd, "*.fsproj") |> Seq.head
  let proj     = Execution.ProjectInstance projFile

  let refs     = proj.GetItems "Reference"
  let projRefs = proj.GetItems "ProjectReference" 
  let files    = proj.GetItems "Compile"

  let prefs   = Seq.map projRefToItem projRefs
  let refs'  = Seq.map (fun s -> s :> ITaskItem) refs
  let prefs' = Seq.map (fun s -> s :> ITaskItem) prefs

  let refDepGraph = mkGraph (Seq.append refs' prefs')
  let refDepOrder = Digraph.topSort refDepGraph
  let refStrings  = List.map spec refDepOrder
  let refNubbed   = nubBy (fun (x,_) -> x) refStrings
  let refFiltered = List.filter (fun (s: String,_) -> List.forall (fun t -> not (s.EndsWith t)) preloads) refNubbed
  let refIncludes = List.map (fun (ref:String,_) -> "#r \"" + ref.Replace(@"\", @"\\") + "\"") refFiltered

  let fsFiles = 
    Seq.map (fun (fsFile:Execution.ProjectItemInstance) -> "#load \"" + (Path.Combine(sd,fsFile.EvaluatedInclude)) + "\"") files

  let script = refIncludes // Seq.append refIncludes fsFiles

  let tempFile = Path.Combine(sd, "__temp__.fsx")
  do File.WriteAllLines(tempFile, script) 
  do List.iter cpDll refNubbed

  do System.Environment.CurrentDirectory <- cd

